#!/usr/bin/env python'

"""Tests for `error_propagation` package"""

import autograd.numpy as np
import numpy.testing
from .arrhenius_model import arrhenius_model, T_expt, k_expt, fitted_params
from ml_uncertainty.non_linear_regression import NonLinearRegression
from ml_uncertainty.model_inference import ParametricModelInference
from scipy.stats import spearmanr as srcc
import pytest

np.random.seed(1)


@pytest.fixture()
def fit_quadratic_model():
    """Quadratic model fitting."""

    def model(X, params):
        return params[0] + X[:, 0] * params[1] + X[:, 0] ** 2 * params[2]

    # Generate X and y data.
    X_expt = np.linspace(0, 10, 100).reshape((-1, 1))
    true_params = np.array([1.0, 1.0, 1.0])
    y_expt = model(X_expt, true_params) + np.random.normal(
        loc=0, scale=1, size=X_expt.shape[0]
    )

    # Non-linear regression.
    nlr = NonLinearRegression(model, p0_length=3, fit_intercept=True)

    nlr.fit(X_expt, y_expt)

    # Predict y from the model.
    y_pred = nlr.predict(X_expt)

    return X_expt, true_params, y_expt, nlr, y_pred


def test_quadratic_model(fit_quadratic_model):
    """Creates a simple non-linear model and tests it."""

    _, true_params, y_expt, nlr, y_pred = fit_quadratic_model

    assert y_pred.shape == y_expt.shape, "Shape of y pred and y expt is different."
    assert (
        nlr.coef_.shape == true_params.shape
    ), "Shape of fitted coefficients \
                    does not match the shape of true coeffients."

    # Check tha that the predicted and true parameters are in the same order
    # of magnitude.
    assert (
        np.linalg.norm(nlr.coef_ - true_params) < 3
    ), "Fitted parameters are different from true parameters."


def test_quadratic_model_inference(fit_quadratic_model):
    """Tests that the model inference of the quadratic model is
    proceeding accurately.
    """

    X_expt, true_params, y_expt, nlr, y_pred = fit_quadratic_model

    inf = ParametricModelInference()

    inf.set_up_model_inference(X_train=X_expt, y_train=y_expt, estimator=nlr)

    # Get prediction intervals on the features.
    df_feature_imp = inf.get_parameter_errors()

    # Getting prediction intervals for X_expt values.
    df_int = inf.get_intervals(X_expt)

    # Check shape of the predicted outputs.
    assert (
        df_feature_imp.shape[0] == true_params.shape[0]
    ), "Feature importance dataframe shape does not match the shape \
            of the true parameters."

    assert (
        df_int.shape[0] == X_expt.shape[0]
    ), "Prediction interval dataframe \
        shape does not match the shape of the input array."

    # Check the predictions made in error_propagation are the same as those made
    # by NLR.
    numpy.testing.assert_almost_equal(
        y_pred,
        df_int["mean"].values,
        decimal=3,
        err_msg="Predicted values from NLR prediction and those from \
                    error_propagation get_intervals method differ from each other.",
    )


def test_arrhenius_model():
    """Tests the non-linear regression for an Arrhenius model.
    We deliberately do not linearize this model so as to try it
    out for a non-linear case.
    """

    # Instantiate the class
    nlr = NonLinearRegression(model=arrhenius_model, p0_length=2)

    # Ensure that the shape of T_expt is (m,1) rather than (m,).
    T_expt_mat = T_expt.reshape((-1, 1))

    # Fit the model and get the parameters.
    nlr.fit(T_expt_mat, k_expt, p0=fitted_params * 0.5)

    k_pred = nlr.predict(T_expt_mat)

    assert k_pred.shape == k_expt.shape, "Shape of k pred and k expt is different."
    assert (
        nlr.coef_.shape == fitted_params.shape
    ), "Shape of fitted coefficients \
                    does not match the shape of true parameters."

    # Check that the ranks of the predicted values and the and the experimental
    # values are the same; i.e., Spearman rank correlation coefficients = 1.
    assert (
        srcc(k_expt, k_pred).correlation == 1
    ), "Predicted value \
        from nlr does not follow the same order as the experimental values. \
        This is unexpected behavior."


def test_arrhenius_model_inference():
    """Test that the non-linear regression function can work with
    parametric model inference to do model inference.
    """

    # Instantiate the class
    nlr = NonLinearRegression(model=arrhenius_model, p0_length=2)

    # Ensure that the shape of T_expt is (m,1) rather than (m,).
    T_expt_mat = T_expt.reshape((-1, 1))

    # Fit the model and get the parameters.
    nlr.fit(T_expt_mat, k_expt, p0=fitted_params * 0.5)

    # Model inference
    inf = ParametricModelInference()

    inf.set_up_model_inference(X_train=T_expt_mat, y_train=k_expt, estimator=nlr)

    df_feature_imp = inf.get_parameter_errors()

    # Compute confidence intervals for new T values.
    T_new = np.linspace(T_expt.min(), T_expt.max(), 50).reshape((-1, 1))

    k_new = nlr.predict(T_new)

    df_int = inf.get_intervals(T_new)

    # Check shape of the predicted outputs.
    assert (
        df_feature_imp.shape[0] == fitted_params.shape[0]
    ), "Feature importance dataframe shape does not match the shape \
            of the fitted parameters."

    assert (
        df_int.shape[0] == T_new.shape[0]
    ), "Prediction interval dataframe \
        shape does not match the shape of the input."

    # Check the predictions made in error_propagation are the same as those made
    # by NLR.
    numpy.testing.assert_almost_equal(
        k_new,
        df_int["mean"].values,
        decimal=3,
        err_msg="Predicted values from NLR prediction and those from \
                    error_propagation get_intervals method differ from each other.",
    )



#!/usr/bin/env python'

"""Tests for `ensemble_model_inference` package"""

import pytest
import pandas as pd
from io import StringIO
import numpy as np
from sklearn.ensemble import RandomForestRegressor
from sklearn.datasets import make_regression
from ml_uncertainty.model_inference.ensemble_model_inference import (
    EnsembleModelInference,
)

np.random.seed(1)


# Fit a regression model.
@pytest.fixture
def model_fit():
    """Sample model fit"""

    X, y = make_regression(
        n_features=4,
        n_informative=2,
        random_state=0,
        shuffle=False,
        n_samples=100,
        n_targets=1,
    )
    regr = RandomForestRegressor(
        max_depth=2, random_state=0, max_samples=0.7, n_estimators=500
    )
    regr.fit(X, y)

    return [X, y, regr]


def test_feature_importance(model_fit):
    """Tests that feature importance can be obtained."""

    X, y, regr = model_fit

    inf = EnsembleModelInference()

    inf.set_up_model_inference(X, y, regr)

    df_imp_list = inf.get_feature_importance_intervals(
        confidence_level=90.0, return_full_distribution=False
    )

    # Getting the variables for dimensions.
    n_features = regr.feature_importances_.shape[0]
    n_outputs = regr.n_outputs_

    # Test the the length of this list is the n_outputs.
    assert (
        len(df_imp_list) == n_outputs
    ), "Length of the feature importance list returned \
                does not equal the number of outputs."

    # Test that the shape of the output df is correct
    assert (
        df_imp_list[0].shape[0] == n_features
    ), "Shape of the dataframe for the first output does not match \
            expected shape. The first dimension of the shape should equal \
            number of features, but that is not the case."

    # Test that the output dataframe is right.
    df_expected_string = ",mean,std,median,lower_bound,upper_bound\n\
        0,0.18222003617429855,0.12723530638421554,0.16748837281064877,0.0,0.3810085909109055\n\
        1,0.8155922827952207,0.1275286651060924,0.8291836478458151,0.6189914090890943,1.0\n\
        2,0.0013591394769462979,0.015180180181922681,0.0,0.0,0.0\n\
        3,0.0008285415535345963,0.008280122731785905,0.0,0.0,0.0\n"
    df_expected = pd.read_csv(StringIO(df_expected_string), index_col=0)

    pd.testing.assert_frame_equal(df_imp_list[0], df_expected, atol=1e-3)



"""Provides commonly used functions for models.
"""
import numpy as np


def linear_model(X, coefs_):
    """
    Linear model without intercept.
    """
    return X @ coefs_


def linear_model_with_intercept(X, coefs_):
    """Linear model where first parameter is the intercept."""
    return coefs_[0] + X @ coefs_[1:]


def ordinary_residual(y_pred, y):
    return y_pred - y


def least_squares_loss(residuals, sample_weight=None):
    r"""
    For ordinary least squares (sample weight is None):
    $$ \mathcal{L}(r, W) = 1/2 r^Tr $$
    For weight least squares (samle weight is not None):
        It must be an array of the shape of residuals.
    $$ \mathcal{L}(r, W) = 1/2 r^TWr $$
    """
    if sample_weight is None:
        loss = 1 / 2 * residuals @ residuals
    else:
        W_mat = np.diag(sample_weight)
        loss = 1 / 2 * residuals @ W_mat @ residuals
    return loss


# More direct way of specifying functions. Can be directly
# defined later.


def linear_residual_function(X, coefs_, y, intercept_=0):
    y_pred = linear_model(X, coefs_, intercept_=intercept_)
    residuals = y_pred - y
    return residuals


def linear_loss_function(X, coefs_, y, intercept_=0):
    """ """
    residuals = linear_residual_function(X, coefs_, y, intercept_=intercept_)
    loss = 1 / (2 * residuals.shape[0]) * (residuals @ residuals)
    return loss


def elastic_net_loss_function(X, coefs_, y, intercept_=0, alpha=1, l1_ratio=0.5):
    """
    Warning: This function doesn't compute the L1
    norm because the error cannot be computed with it.
    """

    residuals = linear_residual_function(X, coefs_, y, intercept_=intercept_)
    loss = (
        1 / (2 * residuals.shape[0]) * (residuals @ residuals)
        + alpha * l1_ratio * np.linalg.norm(coefs_, ord=2)
        + 0.5 * alpha * (1 - l1_ratio) * np.linalg.norm(coefs_, ord=2) ** 2
    )

    return loss



"""Tests that the WLS function works well.

Compares with the example from statsmodels.
Source: https://www.statsmodels.org/dev/examples/notebooks/generated/wls.html

Also see the benchmarking test in
tests/benchmarking/sm_wls.py
"""

import os
import pytest
import autograd.numpy as np
from ml_uncertainty.non_linear_regression import NonLinearRegression
from ml_uncertainty.model_inference import ParametricModelInference
from sklearn.linear_model import LinearRegression


np.random.seed(1024)
file_path = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def set_up_test():

    # Load the expected arrays from the sm outputs.
    path_to_dir = f"{file_path}/benchmarking/sm_wls_outputs"
    x = np.loadtxt(f"{path_to_dir}/x.csv")
    y = np.loadtxt(f"{path_to_dir}/y.csv")
    w = np.loadtxt(f"{path_to_dir}/w.csv")
    sm_params = np.loadtxt(f"{path_to_dir}/sm_params.csv")
    sm_params_se = np.loadtxt(f"{path_to_dir}/sm_params_se.csv")
    sm_pred_df_bounds = np.loadtxt(
        f"{path_to_dir}/sm_pred_df_bounds.csv", delimiter=","
    )

    # In this example, they have used an abridged version on this
    # model for fitting, but they have used the true model for generating y.
    # The model from which y is truly generated is shown in
    # tests/benchmarking/sm_wls.py
    def model(X_arr, coef_):
        """
        $$y = \beta_0 x + \beta_1 x $$
        """
        x = X_arr[:, 0]
        beta0, beta1 = coef_
        y = beta0 + beta1 * x
        return y

    return x, y, w, model, sm_params, sm_params_se, sm_pred_df_bounds


def test_WLS(set_up_test):

    # Set up test
    x, y, w, model, sm_params, sm_params_se, sm_pred_df_bounds = set_up_test

    X_arr = x.reshape((-1, 1))

    # Quantify weights
    weights = 1.0 / w ** 2

    nlr = NonLinearRegression(model=model, p0_length=2, fit_intercept=True)

    nlr.fit(X_arr, y, sample_weight=1.0 / w ** 2)

    inf = ParametricModelInference()

    inf.set_up_model_inference(X_arr, y, nlr, y_train_weights=weights)

    # Get prediction intervals on the features.
    df_feature_imp = inf.get_parameter_errors()

    # Compute prediction intervals.
    df_int = inf.get_intervals(
        X_arr, confidence_level=95.0, distribution="t", y_weights=weights
    )

    # Set up tests to compare with statsmodels
    # Compare sm_best_fit_params with nlr.coef_.
    assert (
        np.linalg.norm(sm_params - inf.best_fit_params) < 1e-1
    ), "Best fit parameters not equal"

    # Compare pred_sm_df with df_feature_imp
    assert (
        np.linalg.norm(sm_params_se - df_feature_imp["std"].values) < 1e-3
    ), "Standard error of parameter values are different"

    # Compare pred_sm_df with df_int
    assert (
        np.linalg.norm(
            (sm_pred_df_bounds - df_int[["lower_bound", "upper_bound"]].values)
            / df_int.shape[0]
        )
        < 1e-1
    ), "Prediction intervals predicted are different"


def test_WLS_with_sklearn(set_up_test):
    """Repeat this example with sklearn OLS fitting."""

    # Set up test
    x, y, w, model, sm_params, sm_params_se, sm_pred_df_bounds = set_up_test

    X_arr = x.reshape((-1, 1))

    # Quantify weights
    weights = 1.0 / w ** 2

    # Fit with sklearn
    regr = LinearRegression(fit_intercept=True)

    regr.fit(X_arr, y, sample_weight=weights)

    inf = ParametricModelInference()

    inf.set_up_model_inference(X_arr, y, regr, y_train_weights=weights)

    df_feature_imp = inf.get_parameter_errors()

    df_int = inf.get_intervals(
        X_arr, confidence_level=95.0, distribution="t", y_weights=weights
    )

    # Set up tests to compare with statsmodels
    # Compare sm_best_fit_params with nlr.coef_.
    assert (
        np.linalg.norm(sm_params - inf.best_fit_params) < 1e-1
    ), "Best fit parameters not equal"

    # Compare pred_sm_df with df_feature_imp
    assert (
        np.linalg.norm(sm_params_se - df_feature_imp["std"].values) < 1e-3
    ), "Standard error of parameter values are different"

    # Compare pred_sm_df with df_int
    assert (
        np.linalg.norm(
            (sm_pred_df_bounds - df_int[["lower_bound", "upper_bound"]].values)
            / df_int.shape[0]
        )
        < 1e-1
    ), "Prediction intervals predicted are different"



"""
Wrapper around scipy.optimize.least_squares to provide easy fitting utility for
non-linear functions.
"""

import numpy as np
from typing import Callable
from copy import deepcopy
from scipy.optimize import least_squares
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.exceptions import DataDimensionalityWarning
from ..error_propagation.error_propagation import ErrorPropagation
from ..model_inference.common_model_functions import ordinary_residual

# Reusing function to check that arrays have the right type and dtype.
check_type_and_dtype = ErrorPropagation.check_type_and_dtype


class NonLinearRegression(RegressorMixin, BaseEstimator):
    r"""
    Scikit-learn-like fitting utility for non-linear models.

    Can be easily used to get model inference such as parameter significance tests
    and prediction intervals.

    Parameters:
    -----------
    model: callable with inputs (X, coefs_, **model_kwargs) and output y
        Function defining how the predictor variables relate to the response.
        X: numpy array of shape (n_examples, n_dimensions)
        coefs_: numpy array of shape (n_parameters, )
        **model_kwargs: additional keyword arguments for the model.
        y: numpy array of shape (n_examples,)
    p0_length: {None, int}, optional
        If None: User must provide an initial guess for the model parameters when
            calling 'fit' method.
        If int: Number of model parameters (n_parameters) including the intercept.
    model_kwargs_dict: dict, optional, default={}
        Dictionary of model keyword arguments.
    residual: {callable, str}, optional, default: "ordinary"
        Defines how to compute residuals.
            If callable (function): It must have its arguments as y_pred, y and
            return residuals. Each of y_pred, y, and residuals must be a 1-D array of
            shape (n_examples,).
            If str: specifies what form of residual function to use.
            As of now, we only support "ordinary" which defines the residual as
            residual = y_hat - y.
    residual_kwargs_dict: dict, optional, default={}
        Dictionary of residual keyword arguments.
    least_sq_kwargs_dict: dict, optional, default={}
        Dictionary of optional keyword arguments to scipy.optimize.least_squares
          function. See scipy.optimize.least_squares documentation.
    copy_X: bool, optional, default=True
        Specified if X should be deep-copied during 'fit', This prevents X input to the
        function from undergoing any change during the fit.
    fit_intercept: bool, optional, default=False
        Whether the first term fit is the model intercept.


    Attributes:
    -----------
    coef_: array of shape (n_parameters, )
        Best parameters of the fit including the intercept where applicable.

    Notes:
    ------
    1. The fit_intercept=True behavior is slightly different from sklearn.
        Here, the self.coef_ variable includes the intercept if it is present.
    2. Only supports single target variable.
    3. Uses only a single processor to fit. Parallelization not implemented.
    4. Handles dense matrices only; no support for sparse matrices.
    5. No regularization available.
    6. Only bounded constraints allowed in scipy.optimize.least_squares are allowed.
        See documentation.

    Examples:
    --------
    examples/non_linear_regression_quadratic.py
    examples/non_linear_regression_arrhenius.py
    """

    def __init__(
        self,
        model: Callable,
        p0_length=None,
        model_kwargs_dict={},
        residual="ordinary",
        residual_kwargs_dict={},
        least_sq_kwargs_dict={},
        copy_X=True,
        fit_intercept=False,
    ):
        self.model = model
        self.p0_length = p0_length
        self.model_kwargs_dict = model_kwargs_dict
        self.least_sq_kwargs_dict = least_sq_kwargs_dict
        self.copy_X = copy_X
        self.fit_intercept = fit_intercept

        # Initialize a function modifying the model as required.
        self.func = self.model

        self._validate_p0_length(self.p0_length)

        # Set residual.
        if residual == "ordinary":
            self.residual = ordinary_residual
        else:
            self.residual = residual

        self.residual_kwargs_dict = residual_kwargs_dict

    def _validate_p0_length(self, p0_length):
        """p0 length must be None of type int"""

        if p0_length is not None:
            if type(p0_length) == int:
                pass
            else:
                raise ValueError(
                    f"p0_length type detected as {type(p0_length)}. \
                                 It must be either None or type int."
                )

    def fit(self, X, y, p0=None, sample_weight=None):
        """
        Fits the non-linear model.

        Wrapper around scipy.optimize.least_squares to implement optimization.

        Parameters:
        -----------
        X: array of shape (n_examples, n_dimensions)
            Training data
        y: array of shape (n_examples, )
            Target values.
        p0: {None, array of shape (n_parameters,)}, optional, default: None
            Array specifying initial estimate of the best fit parameters
            including the intercept.
            If None, an array of shape (p0_length,) of random values between
            0 and 1 from a uniform distribution will be generated and use as the
            initial guess.
        sample_weight: {None, array of shape (n_samples,)}, optional, default: None
            Individual weights for each sample.

        Returns:
        --------
        self: object
            Fitted estimator

        Notes:
        -----
        1. If p0 is None and p0_length is also None, an error will be raised.
        """

        # Validate data
        # Validate X: 2-D array with float dtype.
        if self.copy_X:
            X = deepcopy(X)

        self._validate_X(X)

        # Validate y: 1-D array
        self._validate_y(y, X)

        # Validate p0:
        #   1. If p0 is None, initialize randomly using p0_length: if that is
        # also None, raise error.
        if p0 is None:
            p0 = self._initialize_p0()

        self._validate_p0(p0)

        # Validate sample_weight. Check for the same things as y.
        if sample_weight is not None:
            self._validate_y(sample_weight, X)

        # Validate the model by computing the y_hat for the given X and ensuring
        # that is the same shape as y. Also test the predict method.
        y_pass = self.predict(X, p0, **self.model_kwargs_dict)

        self._validate_y(y_pass, X)

        # Build a residual function.
        def _model_residuals(params, X, y, sample_weight, model_kwargs_dict):

            y_hat = self.predict(X, params, **model_kwargs_dict)
            residuals = self.residual(y_hat, y, **self.residual_kwargs_dict)

            if sample_weight is None:
                return residuals
            else:
                sample_weight_sqrt = np.sqrt(sample_weight)
                return residuals * sample_weight_sqrt

        res_ls = least_squares(
            _model_residuals,
            x0=p0,
            args=(X, y, sample_weight, self.model_kwargs_dict),
            **self.least_sq_kwargs_dict,
        )

        self.fitted_object = res_ls

        self.coef_ = res_ls.x

        if self.fit_intercept:
            self.intercept_ = res_ls.x[0]
        else:
            self.intercept_ = None

    def _validate_X(self, X):
        """Checks that X is an ndarray of 2 dimensions and shape (m, n ).

        Checks that m and n > 0 .
        """

        X = check_type_and_dtype(X)

        if X.ndim == 2 and X.shape[0] > 0 and X.shape[1] > 0:
            pass
        else:
            raise DataDimensionalityWarning(
                "X should be a np.ndarray object with 2 dimensions and shape\
                (m, n) where m>0 and n>0.\
                Please correct the format."
            )

    def _validate_y(self, y, X):
        """Checks that the supplied y_hat is right."""

        # Check type and dtype
        check_type_and_dtype(y)

        # Ensure y_hat has the correct dimension.
        assert y.ndim == 1, "y is not a 1-D vector. Please supply a 1-D vector."
        assert (
            y.shape[0] == X.shape[0]
        ), "Size of y is different from that of X. Please make sure that they match."

    def _initialize_p0(self):
        """"""

        if self.p0_length is not None and type(self.p0_length) == int:
            p0 = np.random.random(self.p0_length)
        else:
            raise ValueError(
                "p0 not supplied and length of p0 could not inferred.\
                                Please supply p0 or supply p0_length value \
                                as an integer."
            )
        return p0

    def _validate_p0(self, p0):
        """Checks that the supplied p0 value is correct."""

        # Check type and dtype
        check_type_and_dtype(p0)

        assert p0.ndim == 1, "p0 is not a 1-D vector. Please supply a 1-D vector."

    def predict(self, X, params=None, **model_kwargs):
        """
        Predict using the given model.

        Parameters:
        -----------
        X: array of shape (n_samples, n_dimensions)
            Samples
        params: {None, array of shape (n_parameters,)}, optional, default=None
            Model parameters.
            If None:
                Will use fitted model parameters.
                If estimator is not fitted, an error will be raised.
        **model_kwargs:
            Optional keyword arguments to be passed to the model.

        Returns:
        --------
        y: array of shape (n_samples,)
            Returns predicted values.
        """

        if params is None:
            params = self.coef_

        if model_kwargs == {}:
            model_kwargs = self.model_kwargs_dict

        return self.func(X, params, **model_kwargs)



"""General utils to be used across functions.
"""


def validate_bool(term, term_name=""):
    """Validates if it is bool."""
    assert type(term) == bool, f"{term_name} is not of type bool."


def validate_str(term, allowed_vals=None, term_name=""):
    """Validates string"""

    assert type(term) == str, f"{term_name} is not type string."

    if allowed_vals is not None:
        assert (
            term in allowed_vals
        ), f"{term_name} is not among allowed values:\
                            {allowed_vals}"


def validate_float(term, term_name=""):
    assert type(term) == float, f"{term_name} is not type float."



"""Top-level package for ML Uncertainty Quantification."""

__author__ = """Archit Nikhil Datar"""
__email__ = "architdatar@gmail.com"
__version__ = "0.1.1"

# Import the required subdirectories.
from . import model_inference
from . import error_propagation
from . import non_linear_regression



"""Parametric model inference class"""

import autograd.numpy as np
from autograd import jacobian
from autograd import elementwise_grad as egrad
import pandas as pd
import sklearn
from sklearn.exceptions import DataDimensionalityWarning
from ..error_propagation.error_propagation import ErrorPropagation
from .common_model_functions import (
    linear_model,
    linear_model_with_intercept,
    ordinary_residual,
    least_squares_loss,
)
from ..error_propagation.statistical_utils import compute_intervals
from ..non_linear_regression import NonLinearRegression

# Reusing function to check that arrays have the right type and dtype.
check_type_and_dtype = ErrorPropagation.check_type_and_dtype


class ParametricModelInference:
    """
    Utility to get model inference for parametric models.

    Can be used to get parameter standard errors and prediction/
    confidence intervals for samples.

    Examples:
    ---------
    examples/non_linear_regression_quadratic.py
    examples/non_linear_regression_arrhenius.py
    examples/parametric_model.py

    References:
    -----------
    1. The functions get_J, get_H are adopted from
        https://github.com/sriki18/adnls/blob/master/adnls.py#L92.
    2. The formulae are taken from Niclas Borlin's lecture slides.
        https://www8.cs.umu.se/kurser/5DA001/HT07/lectures/lsq-handouts.pdf
    """

    def __init__(self):
        """
        Initialize the class
        """

        self.__estimators_implemented = [
            sklearn.linear_model._base.LinearRegression,
            sklearn.linear_model.Lasso,
            sklearn.linear_model.Ridge,
            sklearn.linear_model._coordinate_descent.ElasticNet,
            NonLinearRegression,
        ]

    def set_up_model_inference(
        self,
        X_train,
        y_train,
        estimator=None,
        y_train_weights=None,
        model=None,
        model_kwargs=None,
        intercept=None,
        best_fit_params=None,
        residual=None,
        residual_kwargs=None,
        loss=None,
        loss_kwargs=None,
        regularization=None,
        l1_penalty=None,
        l2_penalty=None,
        custom_reg=None,
        custom_reg_kwargs=None,
        model_dof=None,
    ):
        """Sets up model inference for the fitted model.

        Validates models, corresponding inputs and raises errors if necessary.
        Computes degrees of freedom, resdual mean squared error, and
        the variance-covariance matrix which are important to analyzing the model.

        Parameters
        ----------
        X_train: array of shape (n_training_examples, n_dimensions)
            Training data
        y_train: array of shape (n_training examples, )
            Training targets
        estimator: {None,
                    sklearn.linear_model.LinearRegression,
                    sklearn.linear_model.ElasticNet,
                    ml_uncertainty.non_linear_regression.NonLinearRegression,
                    }, optional, default=None

            If the model is fit through one of the estimators in
                self.__estimators_implemented, the fitted object can be passed
                directly. Several required attributes such as model-model_dof.
                will be inferred automatically.
                Else, these will have to be explicity provided.

            NOTE: Scikit-learn linear models treats the intercept are a different
            parameter from the rest of the model coefficients. It can be fitted
            via the fit_incetercept_=True argument.
            But, for accurate error propagation, the intercept_ must be treated
            as a parameter (as done in statsmodels).

            Thus, for sklearn models, if fit_intercept_=True, we append the
            "intercept_" parameter to the "coefs_" parameter to construct the
            "best_fit_params" and perform error analysis accordingly.
        y_train_weights: array of shape (n_training examples, )
            Weights of training targets. Often used when the variances of the
             distributions from which they are drawn substantially different.
        model: {None, callable}, optional, default=None
            If callable: callable with inputs (X, coefs_, **model_kwargs) and output y
                Function defining how the predictor variables relate to the response.
                X: numpy array of shape (n_examples, n_dimensions)
                coefs_: numpy array of shape (n_parameters, )
                **model_kwargs: additional keyword arguments for the model.
                y: numpy array of shape (n_examples,)
        model_kwargs: {None, dict}, optional, default=None
            If dict: Dictionary of model keyword arguments.
        intercept: {None, float}, optional, default=None
            If the model has an intercept.
        best_fit_params: {None, array}, optional, default: None
            Best fit parameters of the model fit.
            If array:
                Array of shape (n_parameters,).
        residual: {None, callable, str}, optional, default:None
            Defines how to compute residuals.
                If callable (function): It must have its arguments as y_pred, y and
                    return residuals. Each of y_pred, y, and residuals must be a 1-D
                    array of shape (n_examples,).
                If str: specifies what form of residual function to use.
                    As of now, we only support "ordinary" which defines the residual
                    as residual = y_hat - y.
        residual_kwargs: dict, optional, default:None
            Dictionary of residual keyword arguments.
        loss: {None, callable}, optional, default=None
            Defines how loss term in loss function is computed.
            If callable:
                Function with input arguments  residuals (array of shape (n_examples,))
                and other keyword arguments and return float.
                    NOTE: If there is regularization in the loss term, please specify
                    it through the regularization parameters below. Else, the degrees
                    of freedom computation will be incorrect.
        loss_kwargs: {None, dict}, optional, default:None
            Keyword arguments for the loss function.
        regularization: {None, str}, optional, default: None
            Specifies regularization used in the model.
            Used to compute degrees of freedom and variance-covariance matrix.
            If str: specify one of
                "none": No regularization
                "l1": L1 regularization (LASSO)
                "l2": L2 regularization (Ridge)
                "l1+l2": L1 + L2 regularization (elastic net)
                "custom": custom regularization (see custom regulazrization args below)
                    Warning: In this case, please specify model degrees of freedom
                    explicity as that calculation is not implemented here.
        l1_penalty: {None, float}, optional, default:None
            Specifies l1_penalty to be used. See Notes below.
            Only considered if regularization is among {"l1", "l1+l2"}.
        l2_penalty: {None, float}, optional, default:None
            Specifies l2_penalty to be used. See Notes below.
            Only considered if regularization is among {"l2", "l1+l2"}.
        custom_reg: {None, callable}, optional, default: None
            Only considered if regularization=="custom". Else, ignored.
            If callable, function with inputs as (coefs_, **custom_reg_kwargs)
            and outputs a float.
                coefs_: array of shape (n_paramters,) (same as best_fit_params)
        custom_reg_kwargs: {None, dict}, optional, default:None
            Only considered if regularization=="custom". Else, ignored.
            Optional arguments for the function custom_reg.
        model_dof: {None, float}, optional, default: None
            Model degrees of freedom.
            If None: Computed interally using class functions.
            If float: Used directly.

        Notes:
        -------
        1. Loss function is constructed as
                loss = loss_term + L1_penalty * L1_norm(coefs_)
                        + l2_penalty * L2_norm(coefs_)
            If regularization == "custom":
                loss = loss_term + custom_reg(coefs_, **custom_reg_kwargs)
        2. For L1 loss: The derivatives are an approximation.
            Autograd does not differenciate L1 norm, so, for the purposes of
            computing the derivative, we approximate it as L2 norm.
            The errors between the two are small. See benchmarking/L1_v_L2_loss.py
        """

        self.X_train = X_train
        self.y_train = y_train
        self.y_train_weights = y_train_weights
        self.estimator = estimator
        self.model = model
        self.model_kwargs = model_kwargs
        self.intercept = intercept
        self.best_fit_params = best_fit_params
        self.residual = residual
        self.residual_kwargs = residual_kwargs
        self.loss = loss
        self.loss_kwargs = loss_kwargs
        self.regularization = regularization
        self.l1_penalty = l1_penalty
        self.l2_penalty = l2_penalty
        self.custom_reg = custom_reg
        self.custom_reg_kwargs = custom_reg_kwargs
        self.model_dof = model_dof

        # Validate that the estimators, etc. are correctly specified.
        self._validate_input_arguments()

        # If estimator is not None, populate the required model args.
        if self.estimator is not None:
            self._populate_args_for_known_estimators()

        # Compute model and error degrees of freedom.
        self._set_model_dof()
        self._set_error_dof()

        self._set_sigma()  # Square root of RSS

        # Compute parameter standard deviations.
        self._set_parameter_errors()

    def model_function(self, X, coefs_, **model_kwargs):
        return self.model(X, coefs_, **model_kwargs)

    def residual_function(self, X, coefs_, y, model_kwargs, **residual_kwargs):
        """ """
        y_pred = self.model_function(X=X, coefs_=coefs_, **model_kwargs)
        return self.residual(y_pred, y, **residual_kwargs)

    def loss_function(self, X, coefs_, y, model_kwargs, residual_kwargs, **loss_kwargs):
        """ """
        residuals = self.residual_function(
            X, coefs_, y, model_kwargs, **residual_kwargs
        )

        loss_term = self.loss(residuals, **loss_kwargs)

        if self.regularization == "none":
            loss = loss_term
        elif self.regularization == "l1":
            """NOTE: This is an approximation of the L1 norm so that it can
            be differenciated. Autograd does not differenciate L1 norm,
            so this is a stopgap arrangement.
            """
            loss = loss_term + self.l1_penalty * np.linalg.norm(coefs_, ord=2)
        elif self.regularization == "l2":
            loss = loss_term + self.l2_penalty * np.linalg.norm(coefs_, ord=2) ** 2
        elif self.regularization == "l1+l2":
            loss = (
                loss_term
                + self.l1_penalty * np.linalg.norm(coefs_, ord=2)
                + self.l2_penalty * np.linalg.norm(coefs_, ord=2) ** 2
            )
        elif self.regularization == "custom":
            loss = loss_term + self.custom_reg(coefs_, **self.custom_reg_kwargs)
        return loss

    def _validate_input_arguments(self):
        """Validates that the correct input arguments are provided."""

        # Validate X.
        self._validate_X(self.X_train)

        # Validate y.
        self._validate_y(self.y_train)

        # Validate estimator.
        if self.estimator is not None:
            # Check that the estimator is fitted.
            sklearn.utils.validation.check_is_fitted(self.estimator)

            # If estimator is of known type, populate the arguments for it.
            if not type(self.estimator) in self.__estimators_implemented:
                raise ValueError(
                    f"Estimator of type {type(self.estimator).__module__} \
                                 cannot be used to infer the desired properties for \
                                 error analysis. Please supply the required functions \
                                 externally."
                )
        else:
            # Check that all the required functions are provided.
            if any(
                [
                    self.model is None,
                    self.residual is None,
                    self.loss is None,
                    self.best_fit_params is None,
                ]
            ):
                raise ValueError(
                    "At least one of model_function,\
                        residual_function, loss_function, \
                        best fit params, regularization \
                        is not defined.\
                        Please provide all the above as it is necessary for \
                        model inference."
                )

        # Validate y_train_weights
        self._validate_y_train_weights(self.y_train_weights, self.y_train)

        # If custom regularization, make sure that the model DOFs are provided.
        if self.regularization == "custom":
            self._validate_custom_reg(
                self.custom_reg, self.custom_reg_kwargs, self.model_dof
            )
        # Validate model_dof
        self._validate_model_dof(self.model_dof)

    def _validate_X(self, X):
        """Checks that X is an ndarray of 2 dimensions and shape (m, n ).

        Checks that m and n > 0 .
        """

        X = check_type_and_dtype(X)

        if X.ndim == 2 and X.shape[0] > 0 and X.shape[1] > 0:
            return X
        else:
            raise DataDimensionalityWarning(
                "X should be a np.ndarray object with 2 dimensions and shape\
                (m, n) where m>0 and n>0.\
                Please correct the format."
            )

    def _validate_y(self, y):
        """Validates y variable."""

        y = check_type_and_dtype(y)

        if y.ndim != 1:
            raise DataDimensionalityWarning(
                "y should be a 1-D array.\
                    Please provide appropriately. \
                    Support for multiple targets is not available \
                    as of this version. Please treat each target as a separate model."
            )

    def _populate_args_for_known_estimators(self):
        """Populates required arguments for known estimators.

        For instance, If we know that the estimator is a sklearn class,
        we automatically assign it the required functions and get the
        required parameters.
        """

        if type(self.estimator) == sklearn.linear_model._base.LinearRegression:

            self.__populate_sklearn_linear_attributes()

            # Get regularization info.
            self.regularization = "none"
            self.l1_penalty = None
            self.l2_penalty = None

            self.model_kwargs = dict()
            self.residual_kwargs = dict()
            self.loss_kwargs = dict(sample_weight=self.y_train_weights)

        elif type(self.estimator) == sklearn.linear_model.Lasso:
            self.__populate_sklearn_linear_attributes()

            # Get regularization info.
            self.regularization = "l1"
            self.l1_penalty = self.estimator.alpha
            self.l2_penalty = None

            self.model_kwargs = dict()
            self.residual_kwargs = dict()
            self.loss_kwargs = dict(sample_weight=self.y_train_weights)

        elif type(self.estimator) == sklearn.linear_model.Ridge:
            self.__populate_sklearn_linear_attributes()

            self.residual = ordinary_residual
            self.loss = least_squares_loss

            # Get regularization info.
            self.regularization = "l2"
            self.l1_penalty = self.estimator.alpha
            self.l2_penalty = None

            self.model_kwargs = dict()
            self.residual_kwargs = dict()
            self.loss_kwargs = dict(sample_weight=self.y_train_weights)

        elif (
            type(self.estimator) == sklearn.linear_model._coordinate_descent.ElasticNet
        ):
            self.__populate_sklearn_linear_attributes()

            # Get regularization info.
            self.regularization = "l1+l2"
            self.l1_penalty = self.estimator.alpha * self.estimator.l1_ratio
            self.l2_penalty = 0.5 * self.estimator.alpha * (1 - self.estimator.l1_ratio)

            self.model_kwargs = dict()
            self.residual_kwargs = dict()
            self.loss_kwargs = dict(sample_weight=self.y_train_weights)

        elif type(self.estimator) == NonLinearRegression:
            if self.estimator.fit_intercept:
                self.intercept = self.estimator.intercept_
                # In this case, the coef_includes the intercept.
                self.best_fit_params = self.estimator.coef_
            else:
                self.intercept = None
                self.best_fit_params = self.estimator.coef_

            self.model = self.estimator.model
            self.residual = self.estimator.residual
            self.loss = (
                least_squares_loss  # For non-linear, currently, we only support this.
            )

            # Get regularization info.
            self.regularization = "none"
            self.l1_penalty = None
            self.l2_penalty = None

            self.model_kwargs = self.estimator.model_kwargs_dict
            self.residual_kwargs = self.estimator.residual_kwargs_dict
            self.loss_kwargs = dict(sample_weight=self.y_train_weights)

    def __populate_sklearn_linear_attributes(self):
        """Populate for sklearn linear attributes."""

        # If model if fitted with intercept, add intercept to best_fit_params
        # as the first argument.
        if self.estimator.fit_intercept:
            self.intercept = self.estimator.intercept_
            self.best_fit_params = np.concatenate(
                (np.array([self.estimator.intercept_]), self.estimator.coef_)
            )
            self.model = linear_model_with_intercept
        else:
            self.intercept = None
            self.best_fit_params = self.estimator.coef_
            self.model = linear_model

        self.residual = ordinary_residual
        self.loss = least_squares_loss

    def _validate_y_train_weights(self, y_train_weights, y_train):
        """Runs the same validation protocol as y and then checks if they are the same
        dimension.
        """

        if y_train_weights is not None:
            self._validate_y(y_train_weights)

            assert (
                y_train_weights.shape == y_train.shape
            ), "y_train_weights must be None or be the same shape as y_train."

    def _validate_custom_reg(self, custom_reg, custom_reg_kwargs, model_dof):
        """
        1. If regularization is custom: check that the custom reg function
            works correctly.
        2. Check that the model dofs are specified correctly.
        """

        intercept = self.intercept
        best_fit_params = self.best_fit_params

        if intercept is None:  # Intercept wasn't fit separately.
            coef_ = best_fit_params
        else:
            coef_ = best_fit_params[1:]

        # Check that the custom reg function works correctly.
        reg_term = custom_reg(coef_, **custom_reg_kwargs)

        if type(reg_term) != np.float64:
            raise ValueError(
                "custom_reg call did not yield float.\
                             Please provide correct custom_reg function."
            )

        if model_dof is None:
            raise ValueError(
                "Regularization is set to 'custom' but \
                             'model_dof' has not been specified. \
                             Please specify it explicitly."
            )

    def _validate_model_dof(self, model_dof):
        """Either None or float or int"""
        if model_dof is not None:
            if type(model_dof) not in [int, float]:
                raise ValueError(
                    "model_dof does not have correct type.\
                                 Either it should be None or float or int."
                )

    def _set_model_dof(self):
        """Computes model degrees of freedom.

        References:
        1. https://online.stat.psu.edu/stat462/node/131/
        2. https://www.statsmodels.org/dev/generated/statsmodels.
            regression.linear_model.RegressionResults.html#statsmodels.
            regression.linear_model.RegressionResults
        """

        model_dof = self.model_dof
        regularization = self.regularization

        intercept = self.intercept
        best_fit_params = self.best_fit_params

        if model_dof is not None:
            # Model degrees of freedom externally computed and specified.
            return model_dof
        else:  # Compute model degrees of freedom internally.
            if intercept is None:  # Intercept wasn't fit separately.
                coef_ = best_fit_params
            else:
                coef_ = best_fit_params[1:]

            if regularization == "none":
                model_dof = coef_.shape[0]
            elif regularization == "l1":
                non_zero_mask = coef_ != 0
                model_dof = coef_[non_zero_mask].shape[0]
            elif regularization == "l2":
                """Source: Elements of Statistical Learning, Ed. 2, Pg 68"""
                d_values = np.linalg.svd(self.X_train)[1]
                l2_penalty = self.l2_penalty
                model_dof = (d_values ** 2 / (d_values ** 2 + l2_penalty)).sum()
            elif regularization == "l1+l2":
                """Elastic net-type.
                In this case, we refer to the talk by Hui Zou
                (https://hastie.su.domains/TALKS/enet_talk.pdf)
                """
                non_zero_mask = coef_ != 0
                d_values = np.linalg.svd(self.X_train)[1]

                # Since there is L1 regualarization, apply non-zero mask
                # to singular values of X_train.
                d_values = d_values[non_zero_mask]
                l2_penalty = self.l2_penalty
                model_dof = (d_values ** 2 / (d_values ** 2 + l2_penalty)).sum()

        # The intercept value is not penalized during regularization.
        # # So, it should be added as +1 to the model degrees of freedom if it is
        # # part of the model kwargs.
        # if intercept in self.model_kwargs.keys() and intercept is not None:
        #     model_dof += 1

        # Set model dof.
        self.model_dof = model_dof

    def _set_error_dof(self):
        """Error dof = n-p-1 when intercept is present.
        Error dof = n-p when intercept is absent.
        This is in accordance with statsmodels.
        See docs for RegressionResults.
        """
        if self.intercept:
            self.error_dof = self.X_train.shape[0] - 1 - self.model_dof
        else:
            self.error_dof = self.X_train.shape[0] - self.model_dof

    def _set_sigma(self):
        """Computes residual mean squared error of the fit."""

        res = self.residual_function(
            self.X_train,
            self.best_fit_params,
            self.y_train,
            self.model_kwargs,
            **self.residual_kwargs,
        )

        # Also allow for inference from weighted least
        # squares, ultimately allowing for heteroskedasticity in y.
        if self.y_train_weights is not None:
            res = np.sqrt(self.y_train_weights) * res

        sigma = np.sqrt(np.matmul(res.transpose(), res) / self.error_dof)
        self.sigma = sigma

    def get_J(
        self,
        residual_function,
        best_fit_params,
        X,
        y,
        model_kwargs,
        residual_kwargs,
    ) -> np.ndarray:
        """Jacobian of residuals.

        Parameters
        ----------
        residual_function: callable
            Defines how to compute residuals.
            Arguments: (X, best_fit_params, y, model_kwargs, **residual_kwargs)
                and return residuals.
                Each of y, and residuals must be a 1-D
                array of shape (n_examples,).
                The specifications of other input parameters are described below.
        best_fit_params: arrray of shape (n_parameters,)
            Best fit parameters of the model fit.
        X: array of shape (n_samples, n_dimensions)
            Input array
        y: array of shape (n_samples,)
            Targets
        model_kwargs: dict
            Keyword arg dict for the model to relating y to X.
        residual_kwargs:
            Keyword arguments for residual_function.

        Returns
        -------
        J: array of shape (n_samples, n_parameters)
            Jacobian of residuals.
        """

        coefs_ = best_fit_params

        J = jacobian(
            lambda coefs_: residual_function(
                X, coefs_, y, model_kwargs, **residual_kwargs
            )
        )(coefs_)
        return J

    def get_H(
        self,
        loss_function,
        best_fit_params,
        X,
        y,
        model_kwargs,
        residual_kwargs,
        loss_kwargs,
    ) -> np.ndarray:
        """Hessian of objective function.

        Hessian of the objective function is the Jacobian of the gradient
        of the objective funciton.

        Parameters
        ----------
        loss_function: callable with arguments (X, coefs_, y, model_kwargs,
                                                residual_kwargs, loss_kwargs).
                        Returns float
                    X: see below
                    coefs_: same as best_fit_params below
                    y, model_kwargs, model_kwargs, residual_kwargs: see below
        X: array of shape (n_samples, n_dimensions)
            Input array
        best_fit_params: arrray of shape (n_parameters,)
            Best fit parameters of the model fit.
        y: array of shape (n_samples,)
            Targets
        model_kwargs: dict
            Keyword arg dict for the model to relating y to X.
        residual_kwargs: dict
            Keyword arguments for residual_function.
        loss_kwargs:
            Keyword arguments for loss function.

        Returns
        -------
        H: array of shape (n_parameters, n_parameters)
            Hessian of the objective function.
        """

        coefs_ = best_fit_params

        H = jacobian(
            egrad(
                lambda coefs_: loss_function(
                    X,
                    coefs_,
                    y,
                    model_kwargs,
                    residual_kwargs,
                    **loss_kwargs,
                )
            )
        )(coefs_)

        return H

    def get_vcov(self) -> np.ndarray:
        r"""Variance-covariance matrix of parameters.

        Estimate variance-covariance matrix of the provided parameters.
        The formula used is $$ D = \sigma^2 (\nabla^2 f(x^*))^{-1}$$ as described in
        the lecture notes of Niclas Börlin.
        https://www8.cs.umu.se/kurser/5DA001/HT07/lectures/lsq-handouts.pdf

        Returns
        -------
        vcov
            Variance-covariance matrix of the provided parameters.
        """
        sigma = self.sigma
        H = self.H

        # Check if Hessian is invertible else raise error.
        try:
            Hinv = np.linalg.inv(H)
        except Exception:
            raise DataDimensionalityWarning(
                "The computed Hessian is not invertible.\
                    The variance-covariance matrix for the parameters cannot be\
                    computed."
            )

        vcov = (sigma ** 2) * Hinv
        return vcov

    def get_sd_bf(self):
        """Standard deviation of best-fit parameters.

        Get the standard deviation of the best-fit parmeters.

        Returns
        -------
        sd_bf
            Standard deviations of the best-fit parameters.
        """
        vcov = self.vcov
        sd_bf = np.sqrt(np.diag(vcov))
        return sd_bf

    def _set_parameter_errors(self):
        """Compute the standard errors of the parameters."""

        # Compute the Hessian of the loss function.
        # TODO: For linear and ridge models, implement the closed
        # form solution for Hessian.
        self.H = self.get_H(
            self.loss_function,
            self.best_fit_params,
            self.X_train,
            self.y_train,
            self.model_kwargs,
            self.residual_kwargs,
            self.loss_kwargs,
        )

        # Get variance-covariance matrix.
        self.vcov = self.get_vcov()
        self.sd_coef = self.get_sd_bf()

    def get_parameter_errors(
        self,
        distribution="parametric",
        lsa_assumption=False,
        interval_stat=None,
        confidence_level=90.0,
        side="two-sided",
    ):
        """Gets shows parameter standard errors with appropriate intervals.

        Parameters:
        -----------
        distribution: str, optional, {"parametric", "non-parametric"},
                        default: "parametric"
            Distribution of the parameter standard errors.
        lsa_assumption: bool, optional, default=False
            Whether or not to invoke the large sample approximation.
            If True, assumes the distribution to be normal.
            Else, assumes the distribution to be t distribution with error_dof
            degrees of freedom.
        interval_stat: {None, str}, optional, {"normal", "t"}, default: None
            What kind of interval to return. If None: infers from the lsa_assumption
            as described above.
        confidence_level: float, optional, default: 90.0
            Confidence level (0-100) of the interval returned.
        side: str, optional, {"two-sided", "upper", "lower"}, default: "two-sided"
            Defines the type of interval to be computed.
        """

        means_array = self.best_fit_params
        std_array = self.sd_coef

        if distribution == "non-parametric":
            raise NotImplementedError(
                "Non-parametric distribution of \
                of prediction intervals is not yet implemented."
            )
        elif distribution == "parametric":
            # We choose what kind of interval we wish to
            # compute. Thsi is equivalent to the 'distribution' argument in
            # error_propagation. But, we refrain from using that here to
            # avoid confusion. Instead, we use "interval_stat". So,
            # if it is not specifically mentioned, we use 'normal' if LSA is
            # taken to be true, else we use "t" test.
            if interval_stat is None:
                if lsa_assumption:
                    interval_stat = "normal"
                else:
                    interval_stat = "t"

            interval_array = compute_intervals(
                means_array,
                std_array,
                side=side,
                confidence_level=confidence_level,
                distribution=interval_stat,
                dfe=self.error_dof,
            )

        else:
            raise NotImplementedError(
                "'distribution' argument not permissible. \
                    Please provde it as 'parameteric' or\
                    'non-parametric'."
            )

        # Output dictionary.
        param_err_dict = {}
        param_err_dict["mean"] = means_array
        param_err_dict["std"] = std_array
        param_err_dict["lower_bound"] = interval_array[0]
        param_err_dict["upper_bound"] = interval_array[1]

        # Convert into dataframe
        param_err_df = pd.DataFrame.from_dict(param_err_dict)

        return param_err_df

    def get_intervals(
        self,
        X,
        X_err=None,
        X_err_denotes_feature_correlation=False,
        y_weights=None,
        type_="prediction",
        distribution=None,
        lsa_assumption=True,
        confidence_level=90.0,
        side="two-sided",
        dfe=None,
    ):
        r"""Gets intervals for sample data.

        Uses model parameter errors as well as uncertainties in input data
        to compute the desired intervals.

        Parameters:
        -----------
        X: array of shape (n_samples, n_dimensions)
            Input array
        X_err: {None, np.ndarray}, optional, default: None
            Herer, m : number of samples,
            n : number of features
            p: number of parameters

            If X_err is array:
                1. If X_err is a 1-D vector of shape (n,)
                Standard errors for features in X are specified.
                Errors are equal for all examples and are uncorrelated with each other.
                2. If X_err is a 2-D vector of shape (m, n)
                    Standard errors are specified.
                    These are different across samples but assumed to be
                    uncorrelated across features.
                3. If X_err is a 2-D array of shape (n, n)
                    Covariance matrix for errors are specified.
                    Errors are equal for each example but those for different
                    features might be correlated. The $i, j$ th element of matrix
                    represents $cov(X_err_i, X_err_j)$ representing the covariance
                    of the errors $i, j$ th feature of the data.
                4. If X_err is a tensor of shape (m, n, n)
                    For each example, the nxn matrix denotes the covariance matrix of
                    errors in X.
        X_err_denotes_feature_correlation: bool, optional, default: False
            In the rare case that m=n, X_err will computed according to
            case 2 unless this parameter is
            explicity set to True.
        y_weights: {None, array}, optional, default: None
            If array: must have shape (n_samples, )
            Weights of targets. Often used when the variances of the
             distributions from which they are drawn substantially different.
             Used to compute uncertainty of the prediction through
             $\sigma_i = \sigma / weight_i $.
        type_: str, {"confidence", "prediction"}, default: "confidence"
            Defines the type of interval to compute.
        distribution: str, {"normal", "t"}, default: "normal"
            The type of distribution that the desired interval is from.
        lsa_assumption: bool, optional, default=False
            Whether or not to invoke the large sample approximation.
            If True, assumes the distribution to be normal.
            Else, assumes the distribution to be t distribution with error_dof
            degrees of freedom.
        confidence_level: float, default: 90.0
            Percentage of the distribution to be enveloped by the interval.
            Example: A value of 90 means that you wish to compute a 90% interval.
        side: str, "two-sided", "lower", "upper", default: "two-sided"
            Defines the type of interval to be computed.
        dfe: {None, float, int}, optional, default:None
            Error degrees of freedom of the fit.
            Only needed to compute prediction intervals. Ignored for "confidence".
            If None, uses the self.error_dof attribute; i.e., the computed
            error degrees of the model fit.

        Returns:
        --------
        df_int: pandas.DataFrame
            Dataframe consisting of mean (predicted values),
                std (standard error of the interval (prediction / confidence))
                lower_bound, upper_bound: lower, upper bound of the desired intervals
        """

        # Check that the parameter SE have been predicted.
        if not hasattr(self, "sd_coef"):
            raise ValueError(
                "SD coefficients have not been predicted.\
                             Please fit those first using the 'get_paramter_errors' \
                             function."
            )

        if dfe is None:
            dfe = self.error_dof

        eprop = ErrorPropagation()

        df_int = eprop.get_intervals(
            function_=self.model_function,
            X=X,
            params=self.best_fit_params,
            X_err=X_err,
            params_err=self.vcov,
            X_err_denotes_feature_correlation=X_err_denotes_feature_correlation,
            sigma=self.sigma,
            y_weights=y_weights,
            type_=type_,
            side=side,
            confidence_level=confidence_level,
            distribution=distribution,
            lsa_assumption=lsa_assumption,
            dfe=dfe,
            model_kwarg_dict=self.model_kwargs,
        )

        return df_int



r"""
Develops utlities to get model inference for tree-based models in scikit-learn.

# TODO: Modify documentation.
"""

import numpy as np
from warnings import warn
import pandas as pd
from scipy.sparse import issparse
from sklearn.base import is_classifier
from sklearn.exceptions import DataDimensionalityWarning
from sklearn.utils.validation import check_is_fitted
from sklearn.ensemble._forest import (
    _get_n_samples_bootstrap,
    _generate_unsampled_indices,
)
from ..error_propagation.statistical_utils import (
    get_significance_levels,
    get_z_values,
)
from .gen_utils import validate_str, validate_bool, validate_float
from copy import deepcopy


class EnsembleModelInference:
    """Provides utilities to perform inference for ensemble models.

    Examples:
    ---------
    examples/ensemble_model.py
    examples/random_forest_non-normal_distribution.py

    References:
    -----------
    1. Confidence interval calculations are performed using the answer
    provided by @Greg Snow in this Stackoverflow thread on the subject.
    https://stats.stackexchange.com/questions/56895/
    do-the-predictions-of-a-random-forest-model-have-a-prediction-interval
    2. Prediction intervals can be computed using two methods:
        "marginal": This is the 'OOB Prediction Interval' method proposed
        by Zhang et al. (2020).
            DOI: 10.1080/00031305.2019.1585288
        "individual": This is method proposed by @Greg Snow in Ref 1.
        The 'marginal' method is named here because it aggregates the
        predictions from the trees. On the other hand, 'individual' considers
        the distributions of the specific trees.

    """

    def __init__(self):
        """Initialize the function"""

    def set_up_model_inference(
        self,
        X_train,
        y_train,
        estimator,
        use_oob_pred=True,
        variance_type_to_use="marginal",
        copy_X_and_y=True,
        sigma=None,
        distribution="normal",
        distribution_kwargs={},
        stat_generating_function=None,
        stat_generating_function_kwargs={},
    ):
        """Sets up model inference and performs basic calculations.

        Validates estimators and functions, raises errors if anything
        is incorrect. Computes residual mean squared errors which are important
        to getting desired intervals.

        Parameters:
        -----------
        X_train: array of shape (n_training_examples, n_dimensions)
            Training data
        y_train: array of shape (n_training examples, )
            Training targets
        estimator : scikit-learn ensemble model of type RandomForestRegression,
            RandomForestClassification, etc.
            Must be fitted.
        use_oob_pred: bool, optional, default:True
            Whether to use out-of-bag predictions for train data to compute
            model mean squared errors.
        variance_type_to_use: str, optional, {"marginal", "individual"},
                                default="marginal"
            How to compute mean squared errors.
            "marginal": OOB Prediction intervals in Zhang et al. (Ref 2, class doc).
            "individual": Method proposed by Greg Snow (Ref 1, class docs).
        copy_X_and_y: bool, optional, default: True
            Whether to deepcopy X_train and y_train during analysis.
        sigma: {NoneType, float, array}, optional, default: None
            Allows user to externally specify root mean squared error (RMSE)
            for the fit.
            If NoneType: The code computes based on 'variance_type_to_use' argument.
            If float: Value interpreted as the RMSE of the fit.
            If array: Must have shape (n_estimators,), ith element corresponding
                to the RMSE of the ith tree.
        distribution: {NoneType, str, callable}, optional, default: "normal"
            Denotes assumed distribution of the residuals. Should be centered around 0.
            Ignored if variance_type_to_use=='marginal' and 'estimate_from_SD=False'
                in get_intervals.
            If NoneType:
                No distribution assumed. Only works if ignored.
            If str:
                Must be "normal". We provide functions internally to deal with
                normal distribution.
            If callable:
                Provide a function with signature identical to
                numpy.random.normal. See https://numpy.org/doc/stable/reference/random/
                generated/numpy.random.normal.html.
                Also, see ./examples/random_forest_non-normal_distribution.py
                Inputs: (loc, scale, size, **kwargs) and outputs: scalar or array
                scale and size should be able take and return numpy arrays.
        distribution_kwargs: dict, optional, default={}
            Key word arguments for the distribution function.
        stat_generating_function: {NoneType, callable}, optional, default: None
            Function to generate distribution stats for a list of quantile values.
            If NoneType:
                Constructed from 'distribution' using the
                self._stat_generator_function if distribution is not None.
            If callable:
                Must have signature (quantiles, **kwargs) -> list
                Where quantiles are a list and function must return a list of
                distribution statistics.
        stat_generating_function_kwargs: dict, optional, default: {}
            Key word arguments for the stat_generating_function.

        Notes:
        ------
        1.  Set 'use_oob_pred=True' only if there is a large number of trees
            ('n_estimators' > 100), boostrap sampling is used and proportion of samples
            used for each bootstrap is small (<0.8) to ensure that there are enough
            trees for which each sample is out-of-bag and a good distribution can
            be obtained.
        2. The "marginal" method yields prediction intervals for which the coverage
            is closer to the confidence level. So, it is used by default.
            See tests/benchmarking/ensemble_model_prediction_interval.py.
        2. This implementation works for multioutput regression; i.e.,
            y represents more than one target variable (shape (n_samples, n_targets)).
            However, this approach is not recommended. Instead, it is better to fit a
            separate random forest for each target and obtain inference accordingly.
            Reason: The RandomForestRegressor class outputs identical feature
            importances for all target variables leading to identical
            feature importance SD and intervals which is not ideal.
        """

        self.copy_X_and_y = copy_X_and_y
        self.sigma = sigma
        validate_bool(copy_X_and_y)

        self.X_train = self.__validate_and_transform_X(X_train, copy_X_and_y)
        self.y_train = self.__validate_y_and_transform_y(y_train, X_train, copy_X_and_y)

        # Ensure that the estimator is a tree.
        self._confirm_estimator_is_from_ensemble(estimator)
        # Check that the estimator is fitted.
        check_is_fitted(estimator)
        self.estimator = estimator

        validate_bool(use_oob_pred)
        self.use_oob_pred = use_oob_pred

        validate_str(variance_type_to_use, ["marginal", "individual"])
        self.variance_type_to_use = variance_type_to_use

        self.copy_X_and_y = copy_X_and_y

        self._validate_sigma(sigma)

        self._validate_and_assign_distribution_attributes(
            distribution,
            distribution_kwargs,
            stat_generating_function,
            stat_generating_function_kwargs,
        )

        # Further, validate if variance_type_to_use=="individual",
        # "distribution" cannot be None.
        if self.variance_type_to_use == "individual" and self.distribution is None:
            raise ValueError(
                "variance_type_to_use=='individual' \
                             requires that 'distribution' not be None.\
                             However, that value was passed."
            )

        # Validate the distribution and stat generation functions to make
        # sure that they run correctly. Pass a 3D numpy normal array to them and
        # make sure that they spit the right output.
        self.__validate_distribution(self.distribution, self.distribution_kwargs)
        self.__validate_stat_generating_function(
            self.stat_generating_function, self.stat_generating_function_kwargs
        )

        # Get the oob_pred and n_oob_pred arrays.
        train_ind_pred, train_ind_n_pred = self._generate_oob_array(
            X_train, estimator, is_train_data=self.use_oob_pred
        )

        # Compute means.
        train_means = np.nanmean(train_ind_pred, axis=1, keepdims=True)

        # If there are any nan values in train means, we raise a warning.
        if np.isnan(train_means).sum() > 0:
            warn(
                "Out-of-bag (OOB) estimate has been attempted but at least one sample \
                          hasn't been out-of-bag even once. This might lead to errors. \
                          Please reduce the 'max_samples' parameter in estimator or \
                          set 'use_oob_pred' to False."
            )

        if self.sigma is None:
            if self.variance_type_to_use == "marginal":
                # Compute RSS overall. (1, 1, n_outputs)
                marginal_dist = self.y_train - train_means
                MSE = np.nanmean(marginal_dist ** 2, axis=0, keepdims=True)
                std = np.sqrt(MSE)
                self.marginal_dist = marginal_dist

            elif self.variance_type_to_use == "individual":
                y_train_rep = np.tile(self.y_train, (1, self.estimator.n_estimators, 1))
                # Compute MSE by tree. (1, n_estimators, n_outputs)
                MSE = np.nanmean(
                    (train_ind_pred - y_train_rep) ** 2, axis=0, keepdims=True
                )
                std = np.sqrt(MSE)

            self.sigma = std

    def __validate_and_transform_X(self, X, copy_X):
        """Setting the right type of X."""

        if copy_X:
            X = deepcopy(X)

        if type(X) == pd.core.frame.DataFrame:
            X = X.values

        # If in case, ndim of X is 1, we reshape it.
        if X.ndim == 1:
            X = X.reshape((-1, 1))

        # Prediction requires X to be in CSR format
        if issparse(X):
            X = X.tocsr()

        assert X.ndim == 2, DataDimensionalityWarning(
            "X could not be converted to a 2D array"
        )

        return X

    def __validate_y_and_transform_y(self, y, X, copy_y):
        """ """

        if copy_y:
            y = deepcopy(y)

        if y.ndim == 1:
            y = y.reshape((-1, 1, 1))
        elif y.ndim == 2:
            y = y.reshape((y.shape[0], 1, y.shape[1]))
        else:
            raise ValueError("y is not of dimension 1 or 2 as required.")

        assert y.shape[0] == X.shape[0], "Shape of y different from corresponding X."

        return y

    def _confirm_estimator_is_from_ensemble(self, estimator):

        # Checks that the estimator is an ensemble model.
        module = getattr(estimator, "__module__")
        if module != "sklearn.ensemble._forest":
            raise TypeError(
                "Supplied estimator is not of type\
                            sklearn.ensemble._forest.\
                            Please ensure that it is of the right type."
            )

    def _stat_generator(self, significance_levels, **kwargs):
        """Stat generating function: Returns a list of stats given a list of
        significance levels."""

        stats = []
        for level in significance_levels:
            if level is None:
                stats.append(None)
            else:
                stat = np.quantile(self.distribution(size=int(1e6), **kwargs), level)
                stats.append(stat)

        return stats

    def _validate_sigma(self, sigma):
        # Validate sigma
        if sigma is not None:
            if type(sigma) == float:
                pass
            elif type(sigma) == np.ndarray:
                assert sigma.shape == (
                    self.estimator.n_estimators,
                ), f"Shape of sigma provided = {sigma.shape} does not equal the \
                        required shape of ({self.estimator.n_estimators},)."

    def _validate_and_assign_distribution_attributes(
        self,
        distribution,
        distribution_kwargs,
        stat_generating_function,
        stat_generating_function_kwargs,
    ):
        # If distribution is known, assign the correct parameters.
        if type(distribution) == str and distribution == "normal":
            self.distribution = np.random.normal
            self.distribution_kwargs = {}
            self.stat_generating_function = get_z_values
            self.stat_generating_function_kwargs = {}
        elif callable(distribution):
            self.distribution = distribution
            self.distribution_kwargs = distribution_kwargs
            if stat_generating_function is None:
                self.stat_generating_function = self._stat_generator
                self.stat_generating_function_kwargs = distribution_kwargs
            else:
                self.stat_generating_function = stat_generating_function
                self.stat_generating_function_kwargs = stat_generating_function_kwargs
        elif distribution is None:
            self.distribution = distribution
            self.distribution_kwargs = distribution_kwargs

            # In this case, stat_generating function must be None.
            # Else, raise error.
            assert (
                stat_generating_function is None
            ), "If distribution is None,\
                stat generating function must alse be None."

            self.stat_generating_function = stat_generating_function
            self.stat_generating_function_kwargs = stat_generating_function_kwargs
        else:
            raise ValueError(
                "'distribution' must either be None, \
                             function or 'normal', but some other value was passed."
            )

    def __validate_distribution(self, distribution, distribution_kwargs):
        """ """
        # Pass 3D array to scale parameter and make sure that it works.
        try:
            shape = (3, 4, 5)
            dist = distribution(
                size=shape, scale=np.random.random(size=shape), **distribution_kwargs
            )
            assert (
                dist.shape == shape
            ), "Distribution test failed. \
                Does not return the desired shape."
        except Exception:
            raise ValueError(
                "'distribution' validation failed. Please \
                             ensure that the size and scale parameter can take \
                             in 3D arrays like the numpy.random.normal function."
            )

    def __validate_stat_generating_function(
        self, stat_generating_function, stat_generating_function_kwargs
    ):
        """ """
        # Pass list to stat generating function and make sure that we get
        # a list of the same length.
        try:
            levels = [0.05, 0.95]
            stats = stat_generating_function(levels, **stat_generating_function_kwargs)

            assert len(stats) == len(
                levels
            ), "Stat generating function test failed.\
                Did not return the desired size."

        except Exception:
            raise ValueError(
                "Please ensure that the size and scale parameter can take\
                             in 3D arrays like the numpy.random.normal function."
            )

    def get_intervals(
        self,
        X,
        is_train_data=False,
        type_="prediction",
        estimate_from_SD=False,
        confidence_level=90.0,
        side="two-sided",
        lsa_assumption=True,
        return_full_distribution=False,
        copy_X=True,
    ):
        r"""
        Computes prediction / confidence intervals regression model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix.
        is_train_data: bool, optional, default=False
            Indicates if the data is from the training set or not. If True,
            only out-of-bag estimates are used to compute the prediction
            interval.
        type_: str, optional, {"prediction", "confidence"}, default: "prediction"
            Type of the interval to be computed.
                Prediction: Refers to the spread of the predicted value.
                Confidence: Refers to the spread of the mean of the predicted values.
        estimate_from_SD: bool, optional, default: False
            How the interval is computed.
                If False, the appropriate quantiles are returned from the generated
                    distribution.
                If True, the parameters of the generated distribution such as
                    standard deviation (SD) are estimated and the appropriate
                    quantiles are returned.
        confidence_level: float, optional, default: 90.0
            Confidence level of the desired interval.
        side: str, optional, {"two-sided", "lower", "upper"}, default: "two-sided"
            Specifies the type of interval returned.
        lsa_assumption: bool, optional, default: True
            Whether large sample approximation can be considered to hold.
            Only used for confidence intervals if estimate_from_SD==True. Else ignored.
            Generally, valid if number of trees > 30.
            If False, confidence intervals to be estimated using the
            self.stat_generating_function.
        return_full_distribution: bool, optional, default=False
            Specifies whether to return some internal data during computation.
            If True, in addition to the normal output, returns:
                oob_pred:  array of dimensions (n_samples, n_estimators, n_outputs)
                    Provides prediction for each sample by each tree for each target.
                n_oob_pred: array of shape (n_samples, 1, n_outputs)
                    Provides counts for the number of trees for which each
                    sample is OOB.
                means_array: array of shape (n_samples, 1, n_outputs)
                    Mean values of the oob_pred over the estimators.
                std_array: array of shape (n_samples, 1, n_outputs)
                    Standard deviations of the predicted distributions.
        copy_X: bool, optional, default: True
            Whether to copy X so that the original will remain unchanged.

        Returns
        -------
        pred_int_list: List of dataframes of length (n_outputs).
            Each dataframe contains the mean, standard deviation, median,
            and the lower and upper bounds of the desired prediction / confidence
            intervals.
        """

        # Validate values.
        X = self.__validate_and_transform_X(X, copy_X)

        validate_bool(is_train_data, term_name="is_train_data")
        validate_str(
            type_, term_name="type_", allowed_vals=["prediction", "confidence"]
        )
        validate_bool(estimate_from_SD, term_name="estimate_from_SD")
        validate_float(confidence_level, term_name="confidence_level")
        validate_str(
            side, term_name="side", allowed_vals=["two-sided", "lower", "upper"]
        )
        validate_bool(lsa_assumption, term_name="lsa_assumption")
        validate_bool(return_full_distribution, term_name="return_full_distribution")
        validate_bool(copy_X, term_name="copy_X")

        estimator = self.estimator

        if is_classifier(estimator) and hasattr(estimator, "n_classes_"):
            raise ValueError(
                "Prediction intervals can only be computed for regressor models. \
                The model passed is a classifier."
            )

        # Validate others
        if (
            type_ == "confidence"
            and estimate_from_SD
            and not lsa_assumption
            and self.stat_generating_function is None
        ):
            raise ValueError(
                "For confidence intervals to be estimated using \
                             standard deviation, if LSA assumption does not hold,\
                             an appropriate 'stat generating function' must be\
                            supplied."
            )

        if (
            type_ == "prediction"
            and estimate_from_SD
            and self.stat_generating_function is None
        ):
            raise ValueError(
                "For prediction intervals to be computed \
                             standard deviation estimates, an appropriate \
                             'stat_generating_function' is necessary.\
                             Please provide this using 'set_up_model_inference' \
                             or set 'estimate_from_SD' to False."
            )

        # Get the oob_pred and n_oob_pred arrays.
        oob_pred, n_oob_pred = self._generate_oob_array(
            X, estimator, is_train_data=is_train_data
        )

        # Take the mean, median and std along the estimators axis.
        # Shape of these arrays
        # should be (n_samples x n_features)
        means_array = np.nanmean(oob_pred, axis=1, keepdims=True)
        median_array = np.nanmedian(oob_pred, axis=1, keepdims=True)
        std_array = np.nanstd(
            oob_pred, axis=1, ddof=1, keepdims=True
        )  # Sample std. dev.

        # Get significance levels and required percentiles.
        significance_levels = get_significance_levels(confidence_level, side)

        n_samples = X.shape[0]
        n_outputs = estimator.n_outputs_

        if type_ == "confidence":
            if not estimate_from_SD:
                interval_array = self._generate_interval_array_non_parametric(
                    side, significance_levels, n_samples, n_outputs, oob_pred
                )
            else:
                if lsa_assumption:
                    # Assume means to be normally distributed.
                    stat_generating_function = get_z_values
                    stat_generating_function_kwargs = {}
                else:
                    stat_generating_function = self.stat_generating_function
                    stat_generating_function_kwargs = (
                        self.stat_generating_function_kwargs
                    )

                interval_array = self._generate_interval_array_parametric(
                    side,
                    stat_generating_function,
                    stat_generating_function_kwargs,
                    significance_levels,
                    means_array,
                    std_array,
                    n_samples,
                    n_outputs,
                )
        elif type_ == "prediction":
            if self.variance_type_to_use == "marginal":
                if not estimate_from_SD:
                    # If marginal, simply use the distribution.
                    q1 = np.quantile(
                        self.marginal_dist,
                        significance_levels[0],
                        axis=0,
                        keepdims=True,
                    )
                    q2 = np.quantile(
                        self.marginal_dist,
                        significance_levels[1],
                        axis=0,
                        keepdims=True,
                    )

                    q1_rep = np.tile(q1, (n_samples, 1, 1))
                    q2_rep = np.tile(q2, (n_samples, 1, 1))

                    lower_bound = means_array + q1_rep
                    upper_bound = means_array + q2_rep

                    interval_array = np.concatenate((lower_bound, upper_bound), axis=1)
                else:
                    std_array = np.tile(self.sigma, (n_samples, 1, 1))

                    interval_array = self._generate_interval_array_parametric(
                        side,
                        self.stat_generating_function,
                        self.stat_generating_function_kwargs,
                        significance_levels,
                        means_array,
                        std_array,
                        n_samples,
                        n_outputs,
                    )

            elif self.variance_type_to_use == "individual":
                if not estimate_from_SD:
                    # Draw from distribution.
                    noise = np.concatenate(
                        [self.distribution(scale=self.sigma) for _ in range(n_samples)],
                        axis=0,
                    )

                    # Add this noise to the predictions from the trees.
                    pred = oob_pred + noise

                    # Now, get the required quantile from this distribution.
                    interval_array = self._generate_interval_array_non_parametric(
                        side, significance_levels, n_samples, n_outputs, pred
                    )

                else:
                    # sigma dims: (1, n_estimators, n_outputs)
                    # std array: (n_samples, 1, n_outputs)
                    mean_var = np.nanmean(self.sigma ** 2, axis=1, keepdims=True)
                    mean_var_rep = np.tile(mean_var, (n_samples, 1, 1))
                    std_array_pred = np.sqrt(std_array ** 2 + mean_var_rep)

                    interval_array = self._generate_interval_array_parametric(
                        side,
                        self.stat_generating_function,
                        self.stat_generating_function_kwargs,
                        significance_levels,
                        means_array,
                        std_array_pred,
                        n_samples,
                        n_outputs,
                    )

        return self._return_dfs(
            n_outputs,
            means_array,
            std_array,
            median_array,
            interval_array,
            return_full_distribution,
            oob_pred,
            n_oob_pred,
        )

    def _generate_oob_array(self, X, estimator, is_train_data):
        """
        Computes the oob_pred and n_oob_pred arrays.
        The required statistics are then drawn from this array.

        Source: https://github.com/scikit-learn/scikit-learn/blob/
        364c77e047ca08a95862becf40a04fe9d4cd2c98/sklearn/ensemble/_forest.py#L530
            Inspired from the _compute_oob_predictions() function with the
            difference being that that function only provides the
            mean values of the predictions while we can provide prediction
            intervals.

        Returns
        -------
        oob_pred : ndarray of shape (n_samples, n_estimators, n_outputs)

        """

        # Specifying dimensions
        n_samples = X.shape[0]
        n_outputs = estimator.n_outputs_
        n_estimators = estimator.n_estimators

        oob_pred_shape = (n_samples, n_estimators, n_outputs)

        # Initializing oob_pred and n_oob_pred arrays
        oob_pred = np.tile(np.float64(np.nan), oob_pred_shape)
        n_oob_pred = np.zeros((n_samples, 1, n_outputs), dtype=int)

        n_samples_bootstrap = _get_n_samples_bootstrap(
            n_samples,
            estimator.max_samples,
        )

        if is_train_data:
            for estimator_ind, estimator in enumerate(estimator.estimators_):
                unsampled_indices = _generate_unsampled_indices(
                    estimator.random_state,
                    n_samples,
                    n_samples_bootstrap,
                )

                # Should yield output dimensions of n_unsampled_indices x n_outputs
                y_pred = estimator.predict(X[unsampled_indices, :])

                # If y_pred is an array, reshape into a matrix. This can happend for
                # n_output = 1
                if y_pred.ndim == 1:
                    y_pred = y_pred.reshape((-1, 1))

                oob_pred[unsampled_indices, estimator_ind, :] = y_pred
                n_oob_pred[unsampled_indices, 0, :] += 1
        else:
            for estimator_ind, estimator in enumerate(estimator.estimators_):
                # Should yield output dimensions of n_samples x n_outputs
                y_pred = estimator.predict(X)

                # If y_pred is an array, reshape into a matrix. This can happend for
                # n_output = 1
                if y_pred.ndim == 1:
                    y_pred = y_pred.reshape((-1, 1))

                oob_pred[:, estimator_ind, :] = y_pred
                n_oob_pred[:, 0, :] += 1

        # Check if at least one the sample has not been out of the bag even once
        # in the bootstraps used.
        for k in range(n_outputs):
            if (n_oob_pred[..., k] == 0).any():
                warn(
                    "Some inputs do not have OOB scores. This probably means "
                    "too few trees were used to compute any reliable OOB "
                    "estimates.",
                    UserWarning,
                )

        # For samples that dont have any reported OOB values, sets their number to
        # n=1 to avoid problems while calculating means.
        # n_oob_pred[n_oob_pred == 0] = 1

        return oob_pred, n_oob_pred

    def _generate_interval_array_non_parametric(
        self, side, significance_levels, n_samples, n_outputs, oob_pred
    ):
        """ """

        if side == "two-sided":
            # Getting the array of required percentiles.
            # Shape: (n_samples, len(percentiles_required), n_features)
            percentiles = [np.round(sig * 100, 1) for sig in significance_levels]
            interval_array = np.nanpercentile(
                oob_pred, np.array(percentiles), axis=1
            ).swapaxes(0, 1)
        elif side == "upper":
            percentiles = [np.round(significance_levels[0], 1)]
            finite_val_array = np.nanpercentile(
                oob_pred, np.array(percentiles), axis=1
            ).swapaxes(0, 1)
            inf_array = np.tile(np.float64(np.inf), (n_samples, 1, n_outputs))
            interval_array = np.concatenate((finite_val_array, inf_array), axis=1)
        elif side == "lower":
            percentiles = [np.round(significance_levels[1], 1)]
            finite_val_array = np.nanpercentile(
                oob_pred, np.array(percentiles), axis=1
            ).swapaxes(0, 1)
            neg_inf_array = np.tile(np.float64(-np.inf), (n_samples, 1, n_outputs))
            interval_array = np.concatenate((neg_inf_array, finite_val_array), axis=1)

        return interval_array

    def _generate_interval_array_parametric(
        self,
        side,
        stat_generating_function,
        stat_generating_function_kwargs,
        significance_levels,
        means_array,
        std_array,
        n_samples,
        n_outputs,
    ):

        stats = stat_generating_function(
            significance_levels, **stat_generating_function_kwargs
        )
        if side == "two-sided":
            interval_array = np.concatenate(
                (
                    means_array + stats[0] * std_array,
                    means_array + stats[1] * std_array,
                ),
                axis=1,
            )
        elif side == "upper":
            finite_val_array = means_array + stats[0] * std_array
            inf_array = np.tile(np.float64(np.inf), (n_samples, 1, n_outputs))
            interval_array = np.concatenate((finite_val_array, inf_array), axis=1)
        elif side == "lower":
            finite_val_array = means_array + stats[1] * std_array
            neg_inf_array = np.tile(np.float64(-np.inf), (n_samples, 1, n_outputs))
            interval_array = np.concatenate((neg_inf_array, finite_val_array), axis=1)
        return interval_array

    def _return_dfs(
        self,
        n_outputs,
        means_array,
        std_array,
        median_array,
        interval_array,
        return_full_distribution,
        oob_pred=None,
        n_oob_pred=None,
    ):
        """Returns results for feture importance as well as prediction intervals."""
        # Output list of dictionaries.
        pred_int_list = []
        for output_ind in range(n_outputs):
            # Dictionary of the outputs which we will ultimately tranform
            # into a dataframe.
            pred_int_dict = {}
            pred_int_dict["mean"] = means_array[:, 0, output_ind]
            pred_int_dict["std"] = std_array[:, 0, output_ind]
            pred_int_dict["median"] = median_array[:, 0, output_ind]
            pred_int_dict["lower_bound"] = interval_array[:, 0, output_ind]
            pred_int_dict["upper_bound"] = interval_array[:, 1, output_ind]

            # Convert into dataframe
            pred_int_df = pd.DataFrame.from_dict(pred_int_dict)
            pred_int_list.append(pred_int_df)

        # Return the required things.
        if return_full_distribution:
            if oob_pred is None and n_oob_pred is None:
                return pred_int_list, means_array, std_array
            else:
                return pred_int_list, oob_pred, n_oob_pred, means_array, std_array
        else:
            return pred_int_list

    def get_feature_importance_intervals(
        self,
        confidence_level=90.0,
        side="two-sided",
        return_full_distribution=False,
    ):
        r"""
        Computes intervals for the spread of feature importances.

        Parameters
        ----------
        confidence_level: float, optional, default: 90.0
            Confidence level of the interval desired.
        side: str, optional, {"two-sided", "lower", "upper"}, default: "two-sided"
            Specifies if the interval desired is two-sided, upper or lower.
        return_full_distribution: bool, optional, default=False
            Specifies whether to return some internal data during computation.
            If True, in addition to the normal output, returns:
                means_array: array of shape (n_features, 1, n_outputs)
                    Mean values of the feature_importance over the estimators.
                std_array: array of shape (n_samples, 1, n_outputs)
                    Standard deviations of the feauture importances over the
                    estimators.

        Returns
        -------
        feature_imp_int_list: List of dataframes of length (n_outputs).
            Each dataframe contains the mean, standard deviation, median,
            and the lower and upper bounds of the desired feature importance interval.
        """

        # Validates inputs.
        validate_float(confidence_level, term_name="confidence_level")
        validate_str(
            side, term_name="side", allowed_vals=["two-sided", "lower", "upper"]
        )
        validate_bool(return_full_distribution, term_name="return_full_distribution")

        estimator = self.estimator

        (
            feature_imp_array,
            n_feature_imp_array,
        ) = self._generate_feature_importance_array(estimator)

        # Take the mean and std along the estimators axis. Shape of these arrays
        # should be (n_samples x n_features)
        means_array = np.nanmean(feature_imp_array, axis=1, keepdims=True)
        median_array = np.nanmedian(feature_imp_array, axis=1, keepdims=True)

        # Computes prediction standard deviation.
        std_array = np.nanstd(
            feature_imp_array, axis=1, ddof=1, keepdims=True
        )  # Sample std. dev.

        # Get significance levels and required percentiles.
        significance_levels = get_significance_levels(confidence_level, side)

        n_features = estimator.feature_importances_.shape[0]
        n_outputs = estimator.n_outputs_

        # Since we do not know anything about the distributions of the
        # feature importances, we do not assume any distribution and
        # show only the distribution quantiles.
        interval_array = self._generate_interval_array_non_parametric(
            side, significance_levels, n_features, n_outputs, feature_imp_array
        )

        return self._return_dfs(
            n_outputs,
            means_array,
            std_array,
            median_array,
            interval_array,
            return_full_distribution,
        )

        # # Output list of dictionaries.
        # feature_imp_int_list = []
        # for output_ind in range(n_outputs):
        #     # Dictionary of the outputs which we will ultimately tranform
        #     # into a dataframe.
        #     feature_imp_int_dict = {}
        #     feature_imp_int_dict["mean"] = means_array[:, 0, output_ind]
        #     feature_imp_int_dict["std"] = std_array[:, 0, output_ind]
        #     feature_imp_int_dict["median"] = median_array[:, 0, output_ind]
        #     feature_imp_int_dict["lower_bound"] = interval_array[:, 0, output_ind]
        #     feature_imp_int_dict["upper_bound"] = interval_array[:, 1, output_ind]

        #     # Convert into dataframe
        #     feature_imp_int_df = pd.DataFrame.from_dict(feature_imp_int_dict)

        #     feature_imp_int_list.append(feature_imp_int_df)

        # # Return the required things.
        # if return_full_distribution:
        #     return (
        #         feature_imp_int_list,
        #         feature_imp_array,
        #         n_feature_imp_array,
        #         means_array,
        #         std_array,
        #     )
        # else:
        #     return feature_imp_int_list

    def _generate_feature_importance_array(self, estimator):
        """Computes feature importance array."""

        n_features = estimator.feature_importances_.shape[0]
        n_estimators = estimator.n_estimators
        n_outputs = estimator.n_outputs_

        # For each tree, store the feature importances.
        feature_imp_shape = (n_features, n_estimators, n_outputs)

        feature_imp_array = np.tile(np.float64(np.nan), feature_imp_shape)
        n_feature_imp_array = np.zeros((n_features, 1, n_outputs), dtype=int)

        # Compute the array for each tree.
        for estimator_ind, estimator in enumerate(estimator.estimators_):
            # Should yield output dimensions of n_features x n_outputs
            tree_imp = estimator.feature_importances_

            # If y_pred is an array, reshape into a matrix. This can happend for
            # n_output = 1
            if tree_imp.ndim == 1:
                tree_imp = tree_imp.reshape((-1, 1))

            feature_imp_array[:, estimator_ind, :] = tree_imp
            n_feature_imp_array[:, 0, :] += 1

        return feature_imp_array, n_feature_imp_array



"""Benchmark with statsmodels weighted least squares regression.

Source: https://www.statsmodels.org/dev/examples/notebooks/generated/wls.html
"""
#%%
import numpy as np
import statsmodels.api as sm
from ml_uncertainty.non_linear_regression import NonLinearRegression
from ml_uncertainty.model_inference import ParametricModelInference

np.random.seed(1024)

# Statsmodels examples.
# Artificial data: Heteroscedasticity 2 groups

nsample = 50
x = np.linspace(0, 20, nsample)
X = np.column_stack((x, (x - 5) ** 2))
X = sm.add_constant(X)
beta = [5.0, 0.5, -0.01]
sig = 0.5
w = np.ones(nsample)
break_int = nsample * 6 // 10
w[break_int:] = 3
y_true = np.dot(X, beta)
e = np.random.normal(size=nsample)
y = y_true + sig * w * e
X = X[:, [0, 1]]  # The squared term has been left out of the model.

mod_wls = sm.WLS(y, X, weights=1.0 / (w ** 2))
res_wls = mod_wls.fit()
# print(res_wls.summary())
sm_params = res_wls.params
sm_param_se = np.sqrt(np.diag(res_wls.cov_params()))  # parameter error
sm_pred_df = res_wls.get_prediction().summary_frame(alpha=0.05)

# Repeat this analysis with NLR model.


def model_true(X_arr, coef_):
    """
    $$y = \beta_0 x + \beta_1 x + beta_2 (x-5)^2$$
    """
    x = X_arr[:, 0]
    beta0, beta1, beta2 = coef_
    y = beta0 + beta1 * x + beta2 * (x - 5) ** 2
    return y


# In this example, they have used an abridged version on this
# model for fitting, but they have used the true model for generating y.
def model(X_arr, coef_):
    """
    $$y = \beta_0 x + \beta_1 x $$
    """
    x = X_arr[:, 0]
    beta0, beta1 = coef_
    y = beta0 + beta1 * x
    return y


# Fit using NLR
X_arr = x.reshape((-1, 1))
beta_arr = np.array(beta[:-1])
# Quantify weights
weights = 1.0 / w ** 2

nlr = NonLinearRegression(model=model, p0_length=beta_arr.shape[0], fit_intercept=True)

nlr.fit(X_arr, y, sample_weight=1.0 / w ** 2)

inf = ParametricModelInference()

inf.set_up_model_inference(X_arr, y, nlr, y_train_weights=weights)

# Get prediction intervals on the features.
df_feature_imp = inf.get_parameter_errors()

# Compute prediction intervals.
df_int = inf.get_intervals(
    X_arr, confidence_level=95.0, distribution="t", y_weights=weights
)

# Set up tests to compare with statsmodels
# Compare sm_best_fit_params with nlr.coef_.
assert np.linalg.norm(sm_params - nlr.coef_) < 1e-1, "Best fit parameters not equal"

# Compare pred_sm_df with df_feature_imp
assert (
    np.linalg.norm(sm_param_se - df_feature_imp["std"].values) < 1e-3
), "Standard error of parameter values are different"

# Compare pred_sm_df with df_int
assert (
    np.linalg.norm(
        (
            sm_pred_df[["obs_ci_lower", "obs_ci_upper"]].values
            - df_int[["lower_bound", "upper_bound"]].values
        )
        / df_int.shape[0]
    )
    < 1e-1
), "Prediction intervals predicted are different"


#%%
# Write out the inputs and outputs to a file.
# np.savetxt("sm_wls_outputs/x.csv", x, fmt="%.10f", delimiter=",")
# np.savetxt("sm_wls_outputs/y.csv", y, fmt="%.10f", delimiter=",")
# np.savetxt("sm_wls_outputs/w.csv", w, fmt="%.10f", delimiter=",")
# np.savetxt("sm_wls_outputs/sm_params.csv", sm_params, fmt="%.10f", delimiter=",")
# np.savetxt("sm_wls_outputs/sm_params_se.csv", sm_param_se, fmt="%.10f", delimiter=",")
# np.savetxt(
#     "sm_wls_outputs/sm_pred_df_bounds.csv",
#     sm_pred_df[["obs_ci_lower", "obs_ci_upper"]],
#     fmt="%.10f",
#     delimiter=",",
# )

# %%



#!/usr/bin/env python'

"""Tests for `error_propagation` package"""

import pandas as pd
from io import StringIO
from ml_uncertainty.error_propagation.error_propagation import ErrorPropagation
from .arrhenius_model import arrhenius_model, T_expt, fitted_params


def test_error_propagation():
    """Tests that errors can be propagated correctly."""

    T = T_expt.reshape((-1, 1))

    # Some parameters and their standard deviations.
    # For reaction from cyclopropane to propene:
    # https://chem.libretexts.org/Bookshelves/Physical_and_Theoretical_Chemistry_Textbook_Maps/Supplemental_Modules_(Physical_and_Theoretical_Chemistry)/Kinetics/06%3A_Modeling_Reaction_Kinetics/6.02%3A_Temperature_Dependence_of_Reaction_Rates/6.2.03%3A_The_Arrhenius_Law/6.2.3.01%3A_Arrhenius_Equation
    # Calculated from math shown in source. Ea: provided, A: calculated accordingly.
    # A: 1/s, Ea: J/mol
    best_fit_params = fitted_params

    # Let's assume the standard errors in these parameters to be 1% of their
    # value
    best_fit_err = best_fit_params * 1 / 100

    # Supply it an estimate of model RMSE (estimate of $\sigma$).
    sigma_hat = 0.0001

    # Computing the predictions and prediction intervals
    # Initialize the error propagation class
    eprop = ErrorPropagation()

    # Propagate errors
    df_int = eprop.get_intervals(
        arrhenius_model,
        T,
        best_fit_params,
        params_err=best_fit_err,
        sigma=sigma_hat,
        type_="prediction",
        side="two-sided",
        confidence_level=90.0,
        distribution="normal",
    )

    # Test that the output dataframe as expected shape and values.
    assert (
        df_int.shape[0] == T.shape[0]
    ), "Shape of the returned dataframe differs from the expected shape."

    # Assert that the errors found are correct.
    df_expected_string = ",mean,se,lower_bound,upper_bound\n\
    0,0.0001618449625364052,0.00012240067458761856,-3.9486231000346165e-05,0.00036317615607315645\n\
    1,0.0020106823707566135,0.0008322681801027813,0.0006417230361182521,0.0033796417053949743\n\
    2,0.02733807030705305,0.010521113280749972,0.010032378967644147,0.044643761646461944\n\
    3,0.1970242774049084,0.07193191139129122,0.07870681205939108,0.3153417427504257\n"
    df_expected = pd.read_csv(StringIO(df_expected_string), index_col=0)

    pd.testing.assert_frame_equal(df_int, df_expected)



"""Benchmark LASSO and ridge models and custom regularizatio.
"""

import pytest
import autograd.numpy as np
from sklearn.datasets import make_regression
from sklearn.linear_model import Lasso
from ml_uncertainty.model_inference import ParametricModelInference
from ml_uncertainty.model_inference.common_model_functions import (
    ordinary_residual,
    least_squares_loss,
)

# linear_model,
# linear_model_with_intercept,

np.random.seed(1)


def linear_model(X, beta):
    """ """
    return beta[0] + X @ beta[1:]


@pytest.fixture
def fit_sklearn_lasso_model():
    """ """
    X_expt, _ = make_regression(
        n_samples=500, n_features=2, n_informative=2, random_state=1
    )

    # Normalize
    X_expt = (X_expt - X_expt.mean(axis=0)) / X_expt.std(axis=0)
    true_params = np.array([1.0, 1.0, 1.0])

    y_expt = linear_model(X_expt, true_params) + np.random.normal(
        loc=0, scale=1, size=X_expt.shape[0]
    )

    regr = Lasso(alpha=1e-1, fit_intercept=True)

    regr.fit(X_expt, y_expt)

    return X_expt, true_params, y_expt, regr


def test_Lasso_model(fit_sklearn_lasso_model):
    """
    Fits LASSO model and does model analysis.
    """

    # Sklearn Lasso.
    X_expt, true_params, y_expt, regr = fit_sklearn_lasso_model

    # Model inference.
    inf = ParametricModelInference()

    inf.set_up_model_inference(X_expt, y_expt, regr)

    df_feature_imp = inf.get_parameter_errors()

    df_int = inf.get_intervals(X_expt)

    assert len(regr.coef_) == 2, "Shape of coef_ not equal to features."

    assert (
        df_feature_imp.shape[0] == inf.best_fit_params.shape[0]
    ), "Standard parameter shape is wrong."

    assert df_int.shape[0] == X_expt.shape[0], "Prediction interval shape is wrong."


def test_custom_regularization(fit_sklearn_lasso_model):
    """
    Do parametric model inference using custom parameters.
    """

    # Sklearn Lasso.
    X_expt, true_params, y_expt, regr = fit_sklearn_lasso_model

    def l1_reg(coef_, alpha=1.0):
        return alpha * np.linalg.norm(coef_, ord=2)

    # If model if fitted with intercept, add intercept to best_fit_params
    # as the first argument.

    inf = ParametricModelInference()

    inf.set_up_model_inference(
        X_train=X_expt,
        y_train=y_expt,
        model=linear_model,
        model_kwargs={},
        intercept=regr.intercept_,
        best_fit_params=np.concatenate((np.array([regr.intercept_]), regr.coef_)),
        residual=ordinary_residual,
        residual_kwargs={},
        loss=least_squares_loss,
        loss_kwargs={},
        regularization="custom",
        custom_reg=l1_reg,
        custom_reg_kwargs=dict(alpha=1e-1),
        model_dof=2,
    )

    df_feature_imp = inf.get_parameter_errors()

    df_int = inf.get_intervals(X_expt)

    # LASSO model inference.
    inf_Lasso = ParametricModelInference()
    inf_Lasso.set_up_model_inference(X_expt, y_expt, regr)
    df_feature_imp_Lasso = inf_Lasso.get_parameter_errors()
    df_int_Lasso = inf_Lasso.get_intervals(X_expt)

    # Check that this matches with the Lasso implementation
    # from the parametric model inference.
    assert (
        np.linalg.norm(df_feature_imp.values - df_feature_imp_Lasso.values)
        / df_feature_imp.shape[0]
        < 1e-3
    ), "Standard errors \
                between custom regularization and Lasso fit don't match."

    assert (
        np.linalg.norm(df_int.values - df_int_Lasso.values) / df_int.shape[0] < 1e-3
    ), "Confidence intervals \
                between custom regularization and Lasso fit don't match."



"""Benchmark a variety of models against results from statsmodels.

However, we don't use statsmodels here as we do not need the users
to otherwise use statsmodels library.

We have run a script 'benchmarking/sm_linear_models.py' to generate
arrays containing the benchmark values for the set of
data to be with with a linear model.
"""

import pytest
import autograd.numpy as np
from sklearn.datasets import make_regression
from sklearn.linear_model import LinearRegression
from ml_uncertainty.non_linear_regression import NonLinearRegression
from ml_uncertainty.model_inference import ParametricModelInference
import os

np.random.seed(1)
file_path = os.path.dirname(os.path.abspath(__file__))


def linear_model(X, beta):
    """ """
    return X @ beta


@pytest.fixture
def get_sm_data_for_1D_test():
    """Gets sm data for the 1D test."""

    X_expt = np.linspace(0, 10, 1000).reshape((-1, 1))
    true_params = np.array([1.0])

    # Shift the mean by 500 to test that there is no centering effect.
    X_expt += 500

    y_expt = linear_model(X_expt, true_params) + np.random.normal(
        loc=0, scale=1, size=X_expt.shape[0]
    )

    # Load the expected arrays from the sm outputs.
    path_to_dir = f"{file_path}/benchmarking/sm_outputs"
    sm_best_fit_params = np.loadtxt(f"{path_to_dir}/1D_best_fit_params.csv")
    sm_param_errors = np.loadtxt(f"{path_to_dir}/1D_param_errors.csv")
    sm_pred_bounds = np.loadtxt(f"{path_to_dir}/1D_pred_bounds.csv", delimiter=",")

    return (
        X_expt,
        true_params,
        y_expt,
        sm_best_fit_params,
        sm_param_errors,
        sm_pred_bounds,
    )


def test_1D(get_sm_data_for_1D_test):
    """When X is 1D, test that our code matches with statsmodels results."""

    # Get test inputs and outputs.
    (
        X_expt,
        true_params,
        y_expt,
        sm_best_fit_params,
        sm_param_errors,
        sm_pred_bounds,
    ) = get_sm_data_for_1D_test

    # Fit with NLR.
    nlr = NonLinearRegression(
        model=linear_model, p0_length=true_params.shape[0], fit_intercept=False
    )

    nlr.fit(X_expt, y_expt)

    inf = ParametricModelInference()

    inf.set_up_model_inference(X_expt, y_expt, nlr)

    # Get prediction intervals on the features.
    df_feature_imp = inf.get_parameter_errors()

    # Getting prediction intervals for X_expt values.
    df_int = inf.get_intervals(X_expt, confidence_level=95.0, distribution="t")

    # Compare sm_best_fit_params with nlr.coef_
    assert (
        np.linalg.norm(sm_best_fit_params - inf.best_fit_params) < 1e-1
    ), "Best fit parameters not equal"

    # Compare pred_sm_df with df_feature_imp
    assert (
        np.linalg.norm(sm_param_errors - df_feature_imp["std"].values) < 1e-1
    ), "Standard error of parameter values are different"

    # Compare pred_sm_df with df_int
    assert (
        np.linalg.norm(
            (sm_pred_bounds - df_int[["lower_bound", "upper_bound"]].values)
            / df_int.shape[0]
        )
        < 1e-1
    ), "Prediction intervals predicted are different"


@pytest.fixture
def get_sm_data_for_2D_test():
    """Gets inputs and expected outputs for 2D test"""

    X_expt, _ = make_regression(
        n_samples=500, n_features=2, n_informative=2, random_state=1
    )
    true_params = np.array([1.0, 1.0])

    # Means of X shifted by 500 to make sure that the analysis stays
    # valid irrespective of the mean of X.
    X_expt = X_expt + 500

    y_expt = linear_model(X_expt, true_params) + np.random.normal(
        loc=0, scale=1, size=X_expt.shape[0]
    )

    # Load the expected arrays from the sm outputs.
    path_to_dir = f"{file_path}/benchmarking/sm_outputs"
    sm_best_fit_params = np.loadtxt(f"{path_to_dir}/2D_best_fit_params.csv")
    sm_param_errors = np.loadtxt(f"{path_to_dir}/2D_param_errors.csv")
    sm_pred_bounds = np.loadtxt(f"{path_to_dir}/2D_pred_bounds.csv", delimiter=",")

    return (
        X_expt,
        true_params,
        y_expt,
        sm_best_fit_params,
        sm_param_errors,
        sm_pred_bounds,
    )


def test_2D(get_sm_data_for_2D_test):
    """Tests a 2D linear model."""

    (
        X_expt,
        true_params,
        y_expt,
        sm_best_fit_params,
        sm_param_errors,
        sm_pred_bounds,
    ) = get_sm_data_for_2D_test

    # Fit with NLR.
    nlr = NonLinearRegression(model=linear_model, p0_length=true_params.shape[0])

    nlr.fit(X_expt, y_expt)

    inf = ParametricModelInference()

    inf.set_up_model_inference(X_expt, y_expt, nlr)

    # Get prediction intervals on the features.
    df_feature_imp = inf.get_parameter_errors()

    # Getting prediction intervals for X_expt values.
    df_int = inf.get_intervals(X_expt, confidence_level=95.0, distribution="t")

    # Compare sm_best_fit_params with nlr.coef_.
    # There are differences due to the random draw in y.
    assert (
        np.linalg.norm(sm_best_fit_params - nlr.coef_) < 1e-1
    ), "Best fit parameters not equal"

    # Compare pred_sm_df with df_feature_imp
    assert (
        np.linalg.norm(sm_param_errors - df_feature_imp["std"].values) < 1e-1
    ), "Standard error of parameter values are different"

    # Compare pred_sm_df with df_int
    assert (
        np.linalg.norm(
            (sm_pred_bounds - df_int[["lower_bound", "upper_bound"]].values)
            / df_int.shape[0]
        )
        < 1e-1
    ), "Prediction intervals predicted are different"


def test_2D_for_sklearn(get_sm_data_for_2D_test):
    """Repeats the 1D example but benchmarks it with sklearn."""

    # Get test inputs and outputs.
    (
        X_expt,
        true_params,
        y_expt,
        sm_best_fit_params,
        sm_param_errors,
        sm_pred_bounds,
    ) = get_sm_data_for_2D_test

    # Fit model with sklearn.
    regr = LinearRegression(fit_intercept=False)

    regr.fit(X_expt, y_expt)

    inf = ParametricModelInference()

    inf.set_up_model_inference(X_expt, y_expt, regr)

    df_feature_imp = inf.get_parameter_errors()

    df_int = inf.get_intervals(X_expt, confidence_level=95.0, distribution="t")

    # Compare with benchmark results.
    assert (
        np.linalg.norm(sm_best_fit_params - regr.coef_) < 1e-1
    ), "Best fit parameters not equal"

    # Compare pred_sm_df with df_feature_imp
    assert (
        np.linalg.norm(sm_param_errors - df_feature_imp["std"].values) < 1e-1
    ), "Standard error of parameter values are different"

    # Compare pred_sm_df with df_int
    assert (
        np.linalg.norm(
            (sm_pred_bounds - df_int[["lower_bound", "upper_bound"]].values)
            / df_int.shape[0]
        )
        < 1e-1
    ), "Prediction intervals predicted are different"



# Import necessary models
from . import parametric_model_inference, ensemble_model_inference
from . import common_model_functions
from .parametric_model_inference import ParametricModelInference
from .ensemble_model_inference import EnsembleModelInference



"""Benchmarks the various methods for ensemble models.
"""

#%%
import numpy as np
from ml_uncertainty.model_inference import EnsembleModelInference
from sklearn.ensemble import RandomForestRegressor


# Set up regression.
X = (np.random.random(size=10000) * 10 + 5).reshape((-1, 1))

int_for_split = int(0.8 * X.shape[0])
X_train = X[:int_for_split, :]
X_test = X[int_for_split:, :]


def model(X, coef_):
    """ """
    return np.dot(X, coef_)


# Training y
true_params = np.array([1.0])
noise = np.random.normal(loc=0, scale=1, size=X.shape[0])
y = model(X, true_params) + noise

y_train = y[:int_for_split]
y_test = y[int_for_split:]

# Be sure to have a large number of trees (n_estimators)
# and max_samples < 1, if we want to analyze the train OOB features.
regr = RandomForestRegressor(
    max_depth=10, random_state=0, max_samples=0.7, n_estimators=500
)

regr.fit(X_train, y_train)

y_pred_train = regr.predict(X_train)
y_pred_test = regr.predict(X_test)


# Set up function to compute coverage as necessary.
def compute_coverage(df_int, y):
    """Gets coverage from the dataframe."""

    lower_bound = df_int["lower_bound"].values
    upper_bound = df_int["upper_bound"].values

    coverage = ((lower_bound < y) & (upper_bound > y)).sum() / y.shape[0]

    return coverage


# Marginal method.

inf_mar = EnsembleModelInference()

inf_mar.set_up_model_inference(X_train, y_train, regr, variance_type_to_use="marginal")

# Gets prediction intervals NOT using SD.
df_int_list_mar_1 = inf_mar.get_intervals(
    X_train,
    is_train_data=True,
    type_="prediction",
    estimate_from_SD=False,
    confidence_level=90.0,
    return_full_distribution=False,
)

df_int_mar_1 = df_int_list_mar_1[0]
coverage_mar_1 = compute_coverage(df_int_mar_1, y_train)

# Test when prediction  intervals are from SD.
df_int_list_mar_2 = inf_mar.get_intervals(
    X_train,
    is_train_data=True,
    type_="prediction",
    estimate_from_SD=True,
    confidence_level=90.0,
    return_full_distribution=False,
)
df_int_mar_2 = df_int_list_mar_2[0]
coverage_mar_2 = compute_coverage(df_int_mar_2, y_train)

# Individual method.
inf_ind = EnsembleModelInference()

inf_ind.set_up_model_inference(
    X_train, y_train, regr, variance_type_to_use="individual"
)

# Gets prediction intervals NOT using SD.
df_int_list_ind_1 = inf_ind.get_intervals(
    X_train,
    is_train_data=True,
    type_="prediction",
    estimate_from_SD=False,
    confidence_level=90.0,
    return_full_distribution=False,
)
df_int_ind_1 = df_int_list_ind_1[0]
coverage_ind_1 = compute_coverage(df_int_ind_1, y_train)

# Test when prediction  intervals are from SD.
df_int_list_ind_2 = inf_ind.get_intervals(
    X_train,
    is_train_data=True,
    type_="prediction",
    estimate_from_SD=True,
    confidence_level=90.0,
    return_full_distribution=False,
)
df_int_ind_2 = df_int_list_ind_2[0]
coverage_ind_2 = compute_coverage(df_int_ind_2, y_train)

# Print the values
print(f"Coverage: Marginal + without SD = {coverage_mar_1}")
print(f"Coverage: Marginal + with SD = {coverage_mar_2}")
print(f"Coverage: Individual + without SD = {coverage_ind_1}")
print(f"Coverage: Individual + with SD = {coverage_ind_2}")

# Conclusions:
# 1. The coverages for marginal method match the confidence level while
# those obtained through the individual method overestimate them.
# 2. The coverages hardly change much based on whether or not "estimate_with_SD" was
# set to True or not.
# %%



"""
Defines functions for forward propagation of errors.
"""

from typing import Callable
import autograd.numpy as np
import pandas as pd
from autograd import jacobian
from sklearn.exceptions import DataDimensionalityWarning
from .statistical_utils import compute_intervals


class ErrorPropagation:
    r"""
    Performs forward propagation of error for any parametric function
    based on the errors of parameters and its input uncerainties.

    Given $ y = f(\bf{X}, \bf{\beta})  $, computes

    $$ \delta y = \delta f(\bf{X}, \bf{\beta}) $$
    $$ \delta y= \sqrt{(\nabla_{\textbf{X}}f) \delta\textbf{X} (\nabla_{\textbf{X}}f)^T
        + (\nabla_{\bm{\beta}}f) \delta\bm{\beta} (\nabla_{\bm{\beta}}f)^T} $$

    Can be used to compute confidence as well as prediction intervals.

    Confidence intervals ~ Interval of mean of the predicted value.
            $$  \mathrm{Var}( \mathbb{E}(\hat{y} | \mathbf{X}, \bm{\beta})) $$
    Prediction intervals ~ Interval of the predicted value.
            $$ \mathrm{Var}(\hat{y} | \bf{X}, \bm{\beta}) $$

    We know that
            $$ \mathrm{Var}(\hat{y} | \bf{X}, \bm{\beta}) =
            \mathrm{Var}( \mathbb{E}(\hat{y} | \mathbf{X}, \bm{\beta})) +
                \sigma^2 $$
    So, we need some estimate of $\sigma^2$ to calculate prediction intervals.


    Assumptions:
    ------------
    1. All samples in X are i.i.d. (i.e., uncorrelated)

    References:
    -----------
    1.https://www.stat.cmu.edu/~cshalizi/36-220/lecture-11.pdf
    """

    def __init__(self):
        """Initialize the function and set parameters as required."""

    def get_intervals(
        self,
        function_: Callable,
        X,
        params,
        X_err=None,
        params_err=None,
        X_err_denotes_feature_correlation=False,
        sigma=None,
        y_weights=None,
        type_="confidence",
        side="two-sided",
        confidence_level=90.0,
        distribution=None,
        lsa_assumption=False,
        dfe=None,
        model_kwarg_dict={},
    ):
        r"""
        Parameters
        ----------
        function_: callable
            Function with respect to which we want to propagate errors.
            The arguments of the this function MUST be of the form:
            f(X, params, **kwargs) otherwise, errors will be raised.
        X: np.ndarray of shape (m, n)
            m : number of examples,
            n : number of features
        params: np.ndarray of shape (p,)
            The parameters of the function.
            p: Number of parameters
        X_err: {None, array}, optional, default: None
            1. If X_err is a 1-D vector of shape (n,)
             Standard errors for features in X are specified.
             Errors are equal for all examples and are uncorrelated with each other.
            2. If X_err is a 2-D vector of shape (m, n)
                Standard errors are specified.
                These are different across samples but assumed to be uncorrelated
                across features.
            3. If X_err is a 2-D array of shape (n, n)
                Covariance matrix for errors are specified.
                Errors are equal for each example but those for different
                features might be correlated. The $i, j$ th element of matrix
                represents $cov(X_err_i, X_err_j)$ representing the covariance of the
                errors $i, j$ th feature of the data.
            4. If X_err is a tensor of shape (m, n, n)
                For each example, the nxn matrix denotes the covariance matrix of
                errors in X.
        params_err: {None, array}, optional, default: None
            If array-type: shape must be (p,) or (p, p)
            1. If a 1-D vector of shape (p,), treats it as standard error for each
                parameter.
                It will be assumed that the various parameters in $\beta$ are
                mutually uncorrelated.
            2. If a 2-D vector of shape (p, p), it will be assumed that this matrix is
            the variance-covariance matrix for the parameters
            $\mathrm{cov}(\bm{\beta})$.
        X_err_denotes_feature_correlation: bool, optional, default: False
            In the rare case that m=n, X_err will computed according to
            case 2 unless this parameter is
            explicity set to True.
        sigma: {None, float}, default: None
            Estimate of the standard deviation of the distribution of $y$
            at given $\bf{X}, \bm{beta}$.
            Only required if type_=="prediction".

            Mathematically,
            $$ y = f(\bf{X}, \bm{\beta}) + \epsilon $$
            Where $ \epsilon$ follows some distribution given by:
            $$ \epsilon \sim \mathcal{N}(\mu, \sigma^2) $$.
            var_y is an estimate of $\sigma_2$.

            Practically, if this function has been fitted earlier,
            the residual sum of squares (RSS) is a commonly used estimator.

            $$ r = y - \hat{y} $$
            $$ \sigma^2 = MSE = \frac{r^Tr}{m-n} $$

            Alternatively, if there is some prior knowledge / intuition
            about $ \sigma^2 $, it can also be used.
        y_weights: {None, array}, optional, default: None
            If array: must have shape (n_samples, )
            Weights of targets. Often used when the variances of the
             distributions from which they are drawn substantially different.
             Used to compute uncertainty of the prediction through
             $\sigma_i = \sigma / weight_i $.
        type_: str, {"confidence", "prediction"}, default: "confidence"
            Defines the type of interval to compute.
        side: str, "two-sided", "lower", "upper", default: "two-sided"
            Defines the type of interval to be computed.
        confidence_level: float, default: 90.0
            Percentage of the distribution to be enveloped by the interval.
            Example: A value of 90 means that you wish to compute a 90% interval.
        distribution: str, {"normal", "t"}, default: "normal"
            The type of distribution that the desired interval is from.
        lsa_assumption: bool, optional, default=False
            Whether or not to invoke the large sample approximation.
            If True, assumes the distribution to be normal.
            Else, assumes the distribution to be t distribution with error_dof
            degrees of freedom.
        dfe: {None, float, int}, optional, default:None
            Error degrees of freedom of the fit.
            Only needed to compute prediction intervals. Ignored for "confidence".
            If None, uses the self.error_dof attribute; i.e., the computed
            error degrees of the model fit.
        model_kwarg_dict: dict, optional, default:{}
            Keyword arguments to be passed to "function_"

        Returns
        -------
        df_int: pandas.DataFrame
            Dataframe consisting of mean (predicted values),
                std (standard error of the interval (prediction / confidence))
                lower_bound, upper_bound: lower, upper bound of the desired intervals.
        """

        # Validate and transform the inputs
        func, X, params, X_err, params_err = self._validate_and_transform_inputs(
            function_,
            X,
            params,
            X_err=X_err,
            params_err=params_err,
            X_err_denotes_feature_correlation=X_err_denotes_feature_correlation,
            sigma=sigma,
            type_=type_,
            side=side,
            confidence_level=confidence_level,
            distribution=distribution,
            lsa_assumption=lsa_assumption,
            dfe=dfe,
            y_weights=y_weights,
        )

        # Ensure that the function meets the required demands by
        # computing y_hat. If it raises errors, we can simply stop here.
        y_hat = self._compute_y_hat(func, X, params, model_kwarg_dict)

        # Validate y_hat to make sure that it is a 1-D array
        self.__validate_y_hat(y_hat, X)

        # Calculate error propagation.
        SE_on_mean, SE_on_prediction = self._propagate_errors(
            func, X, params, X_err, params_err, model_kwarg_dict, sigma, y_weights
        )

        # Determine appropriate SD and compute interval.
        if type_ == "confidence":
            se = SE_on_mean
        if type_ == "prediction":
            if SE_on_prediction is None:
                raise ValueError(
                    "The prediction error cannot be computed because \
                        'sigma' was not supplied. This parameter is needed to \
                        compute prediction intervals. \
                        Please supply this parameter and recompute.\
                        Or, please request for confidence intervals instead."
                )
            se = SE_on_prediction

        # If LSA assumption is true, we need not compute the t distribution
        # since it required model degrees of freedom which we may not always have.
        # Thus, if distribution is set to "None", we select which one to use
        # considering LSA assumption.
        if distribution is None:
            if lsa_assumption:
                distribution = "normal"
            else:
                distribution = "t"

        # Computes the required interval.
        lower_bound, upper_bound = compute_intervals(
            y_hat,
            se,
            side=side,
            confidence_level=confidence_level,
            distribution=distribution,
            dfe=dfe,
        )

        # Return dataframe.
        df = pd.DataFrame()
        df["mean"] = y_hat
        df["se"] = se
        df["lower_bound"] = lower_bound
        df["upper_bound"] = upper_bound

        return df

    def _validate_and_transform_inputs(
        self,
        function_,
        X,
        params,
        X_err,
        params_err,
        X_err_denotes_feature_correlation,
        sigma,
        type_,
        side,
        confidence_level,
        distribution,
        lsa_assumption,
        dfe,
        y_weights,
    ):
        """Validates and transforms inputs for the function"""

        # Validates and transforms the function.
        # As such, there is nothing to validate, so we simply assign it.
        func = function_

        # Validates and transforms X.
        X = self._validate_X(X)

        # Validates and transforms params.
        params = self._validate_params(params)

        # Validates and transforms params_err and X_err
        params_err = self._validate_and_transform_params_err(params_err)
        X_err = self._validate_and_transform_X_err(X_err, X)

        # Check that at least one of params_err and X_err is not None
        if any([X_err is not None, params_err is not None]):
            pass
        else:
            raise Exception(
                "At least one of X_err, params_err should be supplied\
                for error propagation."
            )

        # Validates X_err_denotes_feature_correlation
        assert isinstance(
            X_err_denotes_feature_correlation, bool
        ), "X_err_denotes_feature_correlation not of type bool. \
                Please supply correct value"

        # Validates sigma
        assert any(
            [sigma is None, isinstance(sigma, float)]
        ), "sigma not of type None or float. \
                Please supply value appropriately."

        # Validates type_
        assert isinstance(
            type_, str
        ), "type_ not of type string. \
                Please supply appropriately."

        # Validates side
        assert isinstance(
            side, str
        ), "side not of type string. \
                Please supply appropriately."

        # Validates confidence_level
        assert isinstance(
            confidence_level, float
        ), "confidence_level not of type float. \
                Please supply appropriately."

        # Validates distribution
        assert any(
            [isinstance(distribution, str), distribution is None]
        ), "distribution not of type string.\
                Please supply appropriately."

        # Validates lsa_assumption
        assert isinstance(
            lsa_assumption, bool
        ), "lsa_assumption not of type bool. \
            Please supply correct value"

        # Validates dfe
        assert any(
            [dfe is None, isinstance(dfe, float), isinstance(dfe, int)]
        ), "dfe is not None, float, or int.\
                   Please supply appropriate values."

        # Validates y_weights
        self._validate_y_weights(y_weights, X)

        return func, X, params, X_err, params_err

    @staticmethod
    def check_type_and_dtype(array):
        """Checks that the given object is a numpy array,
        determines it dtype, and if it is int, changes it to float.
        This avoids problems when using autograd functions.
        """

        if type(array) == np.ndarray:
            if array.dtype.name == "float64":  # This passes the test.
                return array

            else:  # Try to convert it to float 64, else give raise error.
                try:
                    array = array.astype("float64")
                    return array
                except Exception:
                    raise ValueError(
                        "The array could not be converted to\
                        dtype 'float64'. Please try to convert it externally\
                        and ensure that the array.dtype == 'float64'. "
                    )

        else:
            raise ValueError(
                "The array is not of type ndarray. Please supply an\
                array of type ndarray"
            )

    def _validate_params(self, params):
        """Checks that params is a 1-D array or can be cast as one."""

        params = self.check_type_and_dtype(params)

        if params.ndim == 1 and params.shape[0] > 0:
            return params
        else:
            raise DataDimensionalityWarning(
                "params is not passed as a vector of shape (p,).\
                Please correct the format."
            )

    def _validate_X(self, X):
        """Checks that X is an ndarray of 2 dimensions and shape (m, n ).

        Checks that m and n > 0 .
        """

        X = self.check_type_and_dtype(X)

        if X.ndim == 2 and X.shape[0] > 0 and X.shape[1] > 0:
            return X
        else:
            raise DataDimensionalityWarning(
                "X should be a np.ndarray object with 2 dimensions and shape\
                (m, n) where m>0 and n>0.\
                Please correct the format."
            )

    def _validate_and_transform_params_err(self, params_err):
        """
        Check and validate params_err.
        """

        if params_err is not None:  # If it has been defined and is not None.
            params_err = self.check_type_and_dtype(params_err)
            p = params_err.shape[0]
            if params_err.ndim == 1 and p > 0:
                # checks that the params_err is a 1-D non-empty vector.
                # This means that these are the standard errors of the parameters.
                return np.diag(params_err ** 2)

            elif params_err.ndim == 2 and params_err.shape == (p, p):
                # params_err is a pxp matrix specifying the covariance matrix of the
                # parameters.
                return params_err
            else:
                raise DataDimensionalityWarning(
                    "params_err not in one of the required shapes.\
                    Please read the docs."
                )
        else:
            return params_err

    def _validate_and_transform_X_err(self, X_err, X):
        """
        Check and validate X_err
        """

        if X_err is not None:  # If X_err is not None.
            X_err = self.check_type_and_dtype(X_err)
            m, n = X.shape

            if X_err.ndim == 1 and X_err.shape[0] == n:
                # if X_err is a 1-D vector of shape (n,), it means
                # error are the same for all examples and uncorrelated with each other.
                X_err_transformed = np.tile(np.diag(X_err ** 2), (m, 1, 1))
            elif X_err.ndim == 2:
                if X_err.shape == (m, n):
                    X_err_transformed = np.apply_along_axis(
                        np.diag, axis=1, arr=X_err ** 2
                    )

                elif X_err.shape == (n, n):
                    # if X_err is a 2-D matrix with (n,n) dimensions, it means that the
                    # errors are same across all examples, but the features
                    # are correlated and their covariance matrix is specified.
                    X_err_transformed = np.tile(X_err, (m, 1, 1))

                if (
                    X_err.shape(n, n)
                    and m == n
                    and self.X_err_denotes_feature_correlation
                ):
                    # Force the program to treat the matrix as (n,n) rather than (m,n).
                    # In this case, assume that the covariance matrix is specified.
                    X_err_transformed = np.tile(X_err, (m, 1, 1))
            elif X_err.ndim == 3 and X_err.shape == (m, n, n):
                # This means that there are different errors in X for different
                # examples.
                # And assumed that the covariance matrix is specified.
                X_err_transformed = X_err
            else:
                # All other cases
                raise DataDimensionalityWarning(
                    "X_err is not correctly specified. Plesae refer to the docs for \
                    the correct specification"
                )

            return X_err_transformed

        else:  # Simple store it as a None object.
            return X_err

    def _compute_y_hat(self, func, X, params, model_kwarg_dict):
        """Computes y_hat for the X, params, args and kwargs provided."""

        # Get the desired y_hat for the prediction.
        y_hat = func(X, params, **model_kwarg_dict)
        return y_hat

    def __validate_y_hat(self, y_hat, X):
        """Checks that the supplied y_hat is right."""

        # Ensure y_hat has the correct dimension.
        assert y_hat.ndim == 1, "y_hat is not a 1-D vector. Please supply a 1-D vector."
        assert (
            y_hat.shape[0] == X.shape[0]
        ), "Size of y_hat is different from that of X. Please make sure that\
            they match."

    def _validate_y_weights(self, y_weights, X):
        """Checks that y_weights is either NoneType or an array of shape
        (X.shape[0],)"""

        if y_weights is not None:
            assert y_weights.shape == (
                X.shape[0],
            ), "y_weights is not None or array of shape \
                (X.shape[0],). Please provide appropriate values."

    def get_grad_matrix_for_params(self, func, X, params, model_kwarg_dict):

        grad_matrix = jacobian(lambda params: func(X, params, **model_kwarg_dict))(
            params
        )

        m = X.shape[0]
        p = params.shape[0]

        assert grad_matrix.shape == (m, p), DataDimensionalityWarning(
            "Grad matrix for params is of an unexpected dimension."
        )

        return grad_matrix

    def get_grad_matrix_for_X(self, func, X, params, model_kwarg_dict):
        """ """

        m, n = X.shape

        mat = jacobian(lambda X: func(X, params, **model_kwarg_dict))(X)

        # Upon taking the Jacobian,
        # We get a matrix of the shape (m, m, n). This also shows the correlations
        # between various samples. As of now, i dont know how to multiply such a matrix
        # or handle the information about the sample-sample correlation.
        # So, we remove that information and only consider the elements along the
        # diagonal.

        grad_matrix = np.diagonal(mat, axis1=0, axis2=1).T  # grad matrix of shape mxn

        assert grad_matrix.shape == (m, n), DataDimensionalityWarning(
            "Grad matrix for X is of an unexpected dimension."
        )

        return grad_matrix

    def get_variances_based_on_params(self, grad_matrix_for_params, params_err):
        r"""Computes $ \mathrm{var}(E(y|X,\beta)) $ based on $ \beta $ alone."""

        m = grad_matrix_for_params.shape[0]

        # Multiply the two.
        mult = grad_matrix_for_params @ params_err @ grad_matrix_for_params.T

        # Take the diagonal elements
        variances_params = np.diagonal(mult)  # This should have a shape of (m, )

        assert variances_params.shape == (m,), DataDimensionalityWarning(
            "Variances based on params have the wrong dimensions."
        )

        return variances_params

    def get_variances_based_on_X(self, grad_matrix_for_X, X_err):

        r"""Computes the $ \mathrm{var}(E(y|X,\beta)) $ based on $ X $ alone."""

        X = self.X
        m = X.shape[0]

        def compute_y_err_for_x(i):
            """
            Computes the error based on the grad matrix and X_err matrix for each \
            example.
            """

            # grad_X_example: 1xn, X_err_example: nxn, grad_X_example.T: nx1
            # total value: 1x1

            grad_X_example = grad_matrix_for_X[i, :].reshape(1, -1)  # 1xn dimensions
            X_err_example = X_err[i, ...]

            output = grad_X_example @ X_err_example @ grad_X_example.T

            assert output.shape == (1, 1), DataDimensionalityWarning(
                "Variance based on X: wrong shape."
            )

            return output[0][0]

        vfunc = np.vectorize(compute_y_err_for_x)
        variances_X = vfunc([i for i in range(m)])

        return np.array(variances_X)

    def _propagate_errors(
        self, func, X, params, X_err, params_err, model_kwarg_dict, sigma, y_weights
    ):
        """This function will perform the computations to compute
        the required properties which can be accessed as required.
        """

        if X_err is not None:
            grad_matrix_for_X = self.get_grad_matrix_for_X(
                func, X, params, model_kwarg_dict
            )
            variances_X = self.get_variances_based_on_X(grad_matrix_for_X, X_err)
        else:
            variances_X = None

        if params_err is not None:

            grad_matrix_for_params = self.get_grad_matrix_for_params(
                func, X, params, model_kwarg_dict
            )
            variances_params = self.get_variances_based_on_params(
                grad_matrix_for_params, params_err
            )

        else:
            variances_params = None

        # Compute the SE on mean.
        SE_on_mean = self._compute_SE_on_mean(
            X_err, params_err, variances_X, variances_params
        )

        # Get prediction intervals.
        SE_on_prediction = self._compute_SE_on_prediction(SE_on_mean, sigma, y_weights)

        return SE_on_mean, SE_on_prediction

    def _compute_SE_on_mean(self, X_err, params_err, variances_X, variances_params):
        r"""Computes the standard error of the mean prediction SE(E(y|X, \beta))"""

        # Add all variances and take a square root.
        if params_err is not None and X_err is None:  # Only error prop from params.
            SE_on_mean = np.sqrt(variances_params)
        elif params_err is None and X_err is not None:  # Only error prop from X.
            SE_on_mean = np.sqrt(variances_X)
        elif params_err is not None and X_err is not None:  # Both are present.
            SE_on_mean = np.sqrt(variances_params + variances_X)

        return SE_on_mean

    def _compute_SE_on_prediction(self, SE_on_mean, sigma, y_weights):
        r"""Uses SE_on_mean to get SE_on_prediciton.

        $$ SE(y|X, \beta) = \sqrt(SE(E(y|X, \beta))^2 + \sigma^2) $$
        """

        if sigma is not None:
            if y_weights is None:  # Equal weights for all samples.
                SE_on_prediction = np.sqrt((SE_on_mean ** 2 + sigma ** 2))
            if y_weights is not None:
                SE_on_prediction = np.sqrt((SE_on_mean ** 2 + sigma ** 2 / y_weights))

        else:
            SE_on_prediction = None

        return SE_on_prediction



"""Tests ensemble model for more complex cases:
specifically 3 cases:
1. Multioutput linear
2. Single output non-linear
3. Errors from non-linear distribution
"""

import pytest
import numpy as np
from ml_uncertainty.model_inference import EnsembleModelInference
from sklearn.datasets import make_regression
from sklearn.ensemble import RandomForestRegressor

np.random.seed(1)


@pytest.fixture
def set_up_multioutput_linear():
    """ """

    X, y = make_regression(
        n_features=4,
        n_informative=2,
        random_state=0,
        shuffle=False,
        n_samples=10000,
        noise=1,
        n_targets=3,
    )
    regr = RandomForestRegressor(
        max_depth=2, random_state=0, max_samples=0.7, n_estimators=500
    )
    regr.fit(X, y)

    return [X, y, regr]


def compute_coverage(df_int, y):
    """Gets coverage from the dataframe."""

    lower_bound = df_int["lower_bound"].values
    upper_bound = df_int["upper_bound"].values

    coverage = ((lower_bound < y) & (upper_bound > y)).sum() / y.shape[0]

    return coverage


def test_multioutput_linear(set_up_multioutput_linear):
    """Tests that the model works well on
    multioutput cases.
    """

    X, y, regr = set_up_multioutput_linear

    inf = EnsembleModelInference()

    inf.set_up_model_inference(X, y, regr)

    # Feature importances
    df_feature_imp_list = inf.get_feature_importance_intervals(
        confidence_level=90.0, side="two-sided"
    )

    # Intervals
    df_int_list = inf.get_intervals(
        X,
        is_train_data=True,
        type_="prediction",
        confidence_level=90.0,
    )
    df_int_0 = df_int_list[0]
    coverage_0 = compute_coverage(df_int_0, y[:, 0])

    df_int_1 = df_int_list[1]
    coverage_1 = compute_coverage(df_int_1, y[:, 1])

    df_int_2 = df_int_list[2]
    coverage_2 = compute_coverage(df_int_2, y[:, 2])

    # Test that the computed coverages are close to the confidence interval.
    assert (
        len(df_feature_imp_list) == y.shape[1]
    ), "Feature importance list length \
        differs from the number of targets."

    assert (
        np.abs(coverage_0 - 0.90) < 0.02
    ), "Coverage for first variable\
        deviates from expected value."

    assert (
        np.abs(coverage_1 - 0.90) < 0.02
    ), "Coverage for second variable\
        deviates from expected value."

    assert (
        np.abs(coverage_2 - 0.90) < 0.02
    ), "Coverage for third variable\
        deviates from expected value."


@pytest.fixture()
def set_up_non_linear_regression():
    """Sets up non-linear regression."""

    X = np.linspace(-10, 10, 10000).reshape((-1, 1))

    def sigmoid(X):
        x = X[:, 0]
        return 1 / (1 + np.exp(-x))

    y_true = sigmoid(X)
    noise = np.random.normal(scale=0.1, size=X.shape[0])

    y = y_true + noise

    # Split into train and test.
    train_idx = np.random.choice(
        np.arange(X.shape[0]), size=int(0.8 * X.shape[0]), replace=False
    )

    X_train = X[train_idx, :]
    y_train = y[train_idx]
    mask = np.ones(X.shape[0], dtype=bool)
    mask[train_idx] = False
    X_test = X[mask, :]
    y_test = y[mask]

    regr = RandomForestRegressor(
        max_depth=10, random_state=0, max_samples=0.7, n_estimators=500
    )

    regr.fit(X, y)

    return X_train, y_train, X_test, y_test, regr


def test_non_linear_function(set_up_non_linear_regression):
    """Tests non-linear function."""

    X_train, y_train, X_test, y_test, regr = set_up_non_linear_regression

    inf = EnsembleModelInference()

    inf.set_up_model_inference(X_train, y_train, regr)

    # Get intervals
    df_int_train = inf.get_intervals(
        X_train, is_train_data=True, type_="prediction", confidence_level=90.0
    )[0]

    df_int_test = inf.get_intervals(
        X_test, is_train_data=False, type_="prediction", confidence_level=90.0
    )[0]

    coverage_train = compute_coverage(df_int_train, y_train)
    coverage_test = compute_coverage(df_int_test, y_test)

    # Tests coverage
    assert (
        np.abs(coverage_train - 0.90) < 0.02
    ), "Coverage for train\
            deviates from confidence_level."

    # The coverage for test set is different than that of the train set.
    assert (
        np.abs(coverage_test - 0.90) < 0.02
    ), "Coverage for test\
            deviates from confidence_level."


@pytest.fixture()
def set_up_non_normal_dist():
    """Sets up regression with non-normal distribution of errors."""

    X = np.linspace(-10, 10, 10000).reshape((-1, 1))

    def linear_model(X, coef_):
        return coef_[0] + np.dot(X, coef_[1:])

    true_params = np.array([1.0, 1.0])
    y_true = linear_model(X, true_params)

    # Poisson distribution.
    noise = np.random.poisson(lam=1, size=X.shape[0]) - 1

    y = y_true + noise

    # Split into train and test.
    train_idx = np.random.choice(
        np.arange(X.shape[0]), size=int(0.8 * X.shape[0]), replace=False
    )

    X_train = X[train_idx, :]
    y_train = y[train_idx]
    mask = np.ones(X.shape[0], dtype=bool)
    mask[train_idx] = False
    X_test = X[mask, :]
    y_test = y[mask]

    regr = RandomForestRegressor(
        max_depth=10, random_state=0, max_samples=0.7, n_estimators=500
    )

    regr.fit(X, y)

    return X_train, y_train, X_test, y_test, regr, noise, true_params


def test_non_normal_error_distribution_marginal(set_up_non_normal_dist):
    """Tests inference for regression with non-normal error distributions."""

    X_train, y_train, X_test, y_test, regr, noise, true_params = set_up_non_normal_dist

    inf = EnsembleModelInference()

    def poisson(scale=1, size=None):
        r"""Wrapper around numpy's Poisson function to specify scale.
        Note: For poisson, the mean, and variance are both=$\lambda$.
        So, we adjust for the mean, and "scale", i.e., standard deviation.
        """

        lam = scale ** 2
        return np.random.poisson(lam=lam, size=size) - lam

    inf.set_up_model_inference(X_train, y_train, regr, distribution=poisson)

    df_int_train = inf.get_intervals(
        X_train, is_train_data=True, confidence_level=90.0
    )[0]
    coverage_train = compute_coverage(df_int_train, y_train)

    df_int_test = inf.get_intervals(X_test, is_train_data=False, confidence_level=90.0)[
        0
    ]
    coverage_test = compute_coverage(df_int_test, y_test)

    # Tests coverage
    assert (
        np.abs(coverage_train - 0.90) < 0.05
    ), "Coverage for train\
            deviates from confidence_level."

    # The coverage for test set is different than that of the train set.
    assert (
        np.abs(coverage_test - 0.90) < 0.05
    ), "Coverage for test\
            deviates from confidence_level."


def test_non_normal_error_distribution_ind(set_up_non_normal_dist):
    """Tests inference for regression with non-normal error distributions.
    Tests this with 'individual' error predictions.
    """

    X_train, y_train, X_test, y_test, regr, noise, true_params = set_up_non_normal_dist

    inf = EnsembleModelInference()

    def poisson(scale=1, size=None):
        r"""Wrapper around numpy's Poisson function to specify scale.
        Note: For poisson, the mean, and variance are both=$\lambda$.
        So, we adjust for the mean, and "scale", i.e., standard deviation.
        """

        lam = scale ** 2
        return np.random.poisson(lam=lam, size=size) - lam

    inf.set_up_model_inference(
        X_train, y_train, regr, distribution=poisson, variance_type_to_use="individual"
    )

    df_int_train = inf.get_intervals(
        X_train, is_train_data=True, confidence_level=90.0
    )[0]
    coverage_train = compute_coverage(df_int_train, y_train)

    df_int_test = inf.get_intervals(X_test, is_train_data=False, confidence_level=90.0)[
        0
    ]
    coverage_test = compute_coverage(df_int_test, y_test)

    # Tests coverage: In this case, it is different
    # from the 90% since this prediction interval has more coverage
    # than confidence level.
    assert (
        np.abs(coverage_train - 0.99) < 0.02
    ), "Coverage for train\
            deviates from confidence_level."

    # The coverage for test set is different than that of the train set.
    assert (
        np.abs(coverage_test - 0.99) < 0.02
    ), "Coverage for test\
            deviates from confidence_level."



"""Describes the arrhenius model as a function which is useful
in several tests.
"""

# Regular numpy doesn't work for error propagation since there are
# issues with respect to taking a derivative.
import autograd.numpy as np


def arrhenius_model(T, coefs_):
    r"""Arrhenius model function

    $$ k = Ae^{-Ea/RT}$$

    Parameters:
    -----------
    T: np.ndarray of dimension 1
        Temperatures in $\degree C$
    coefs_: np.ndarray of shape (2,)
        Denotes [A, Ea].
        A (units: same as those of rate constant k),
        Ea (units: J/mol).

    Returns:
    --------
    np.ndarray of dimension 1.
        Corresponding to k values for each T value.
    """

    X = T[:, 0] + 273  # transforming to K

    R = 8.314  # J/mol/K
    A, Ea = coefs_

    k = A * np.exp(-Ea / (R * X))

    k = k.reshape((-1,))

    return k


# For reaction from cyclopropane to propene T, k, and best fit parameter
# values are shown below.
# https://chem.libretexts.org/Bookshelves/
# Physical_and_Theoretical_Chemistry_Textbook_Maps/
# Supplemental_Modules_(Physical_and_Theoretical_Chemistry)/
# Kinetics/06%3A_Modeling_Reaction_Kinetics/
# 6.02%3A_Temperature_Dependence_of_Reaction_Rates/
# 6.2.03%3A_The_Arrhenius_Law/6.2.3.01%3A_Arrhenius_Equation

# Temperature values ($\degree C$)
T_expt = np.array([477, 523, 577, 623])

# Rate constant (k) values in 1/s
k_expt = np.array([0.0018, 0.0027, 0.030, 0.26])

# Calculated from math shown in source. Ea: provided, A: calculated accordingly.
# A: 1/s, Ea: J/mol
fitted_params = np.array([1.39406453358858e15, 271.867e3])



"""Benchmark with OLS linear model but non-linear curve.
Source website: https://www.statsmodels.org/dev/examples/notebooks/generated/ols.html

NOTE: This benchmarking example has not been included as a test.
This is because it is the exact same concept as the sm_linear_models.py case.
"""

#%%
# Statsmodels example: OLS non-linear curve but linear in parameters.

import autograd.numpy as np
import statsmodels.api as sm
from ml_uncertainty.non_linear_regression import NonLinearRegression
from ml_uncertainty.model_inference import ParametricModelInference

np.random.seed(9876789)

# Statsmodels example
nsample = 50
sig = 0.5
x = np.linspace(0, 20, nsample)
X = np.column_stack((x, np.sin(x), (x - 5) ** 2, np.ones(nsample)))
beta = [0.5, 0.5, -0.02, 5.0]

y_true = np.dot(X, beta)
y = y_true + sig * np.random.normal(size=nsample)

res = sm.OLS(y, X).fit()
# print(res.summary())

# Get prediction intervals.
pred_ols = res.get_prediction()

sm_params = res.params
sm_param_se = np.sqrt(np.diag(res.cov_params()))  # parameter error
sm_pred_df = pred_ols.summary_frame()  # alpha=0.05 (default)

# # Plot
# iv_l = pred_ols.summary_frame()["obs_ci_lower"]
# iv_u = pred_ols.summary_frame()["obs_ci_upper"]

# fig, ax = plt.subplots(figsize=(8, 6))

# ax.plot(x, y, "o", label="data")
# ax.plot(x, y_true, "b-", label="True")
# ax.plot(x, res.fittedvalues, "r--.", label="OLS")
# ax.plot(x, iv_u, "r--")
# ax.plot(x, iv_l, "r--")
# ax.legend(loc="best")

# Recreate this with NLR.


def model(X_arr, coefs_):
    r"""
    $$ y = \beta_0 x + \beta_1 \sin(x) + \beta_2(x-5)^2 + \beta_3
    """
    x = X_arr[:, 0]
    beta0, beta1, beta2, beta3 = coefs_
    y = beta0 * x + beta1 * np.sin(x) + beta2 * (x - 5) ** 2 + beta3
    return y


X_arr = x.reshape((-1, 1))  # Must be 2D array.
beta_arr = np.array(beta)
y_model = model(X_arr, np.array(beta_arr))

# Fit using NLR
nlr = NonLinearRegression(model=model, p0_length=beta_arr.shape[0], fit_intercept=True)

nlr.fit(X_arr, y)

inf = ParametricModelInference()

inf.set_up_model_inference(X_arr, y, nlr)

# Get prediction intervals on the features.
df_feature_imp = inf.get_parameter_errors(confidence_level=95.0)

# Get prediction intervals.
df_int = inf.get_intervals(X_arr, confidence_level=95.0, distribution="t")

# Compare sm_best_fit_params with nlr.coef_
assert np.linalg.norm(sm_params - nlr.coef_) < 1e-3, "Best fit parameters not equal"

# Compare pred_sm_df with df_feature_imp
assert (
    np.linalg.norm(sm_param_se - df_feature_imp["std"].values) < 1e-3
), "Stanard error of parameter values are different"

# Compare pred_sm_df with df_int
assert (
    np.linalg.norm(
        (
            sm_pred_df[["obs_ci_lower", "obs_ci_upper"]]
            - df_int[["lower_bound", "upper_bound"]].values
        )
        / df_int.shape[0]
    )
    < 1e-3
), "Prediction intervals predicted are different"


# %%



#!/usr/bin/env python'

"""Tests for `ensemble_model_inference` package"""

import pytest
import numpy as np
from ml_uncertainty.model_inference.parametric_model_inference import (
    ParametricModelInference,
)


@pytest.fixture
def model_fit():
    """Sample model fit: Elastic net"""

    from sklearn.datasets import make_regression
    from sklearn.linear_model import ElasticNet

    np.random.seed(1)

    # Create a test case for elastic net regression and test the inference with
    # the created class.
    X, y = make_regression(n_samples=20, n_features=2, n_informative=2, noise=1)

    regr = ElasticNet(alpha=0.010, l1_ratio=0.5)

    regr.fit(X, y)

    return X, y, regr


def test_parameter_errors(model_fit):
    """Tests that parameter errors can be computed properly."""

    X, y, regr = model_fit

    inf = ParametricModelInference()

    inf.set_up_model_inference(X_train=X, y_train=y, estimator=regr)

    # Should return a dataframe of parameters and their CIs.
    # Each element corresponding to one output.
    df_imp = inf.get_parameter_errors()

    # Make sure that it is the right shape.
    assert (
        df_imp.shape[0] == regr.coef_.shape[0] + 1  # +1 due to the intercept term.
    ), "First dimension of the feature importance dataframe \
            differs from the number of coefficients in the model."


def test_confidence_intervals(model_fit):
    """Check if model can compute confidence intervals."""

    X, y, regr = model_fit

    inf = ParametricModelInference()

    inf.set_up_model_inference(X_train=X, y_train=y, estimator=regr)

    # Compute confidence intervals.
    df_int = inf.get_intervals(X)

    assert (
        df_int.shape[0] == X.shape[0]
    ), "Shape of the returned interval dataframe does not match the \
            number of samples."



"""
To get Hessian matrix of loss function, we need to take its derivative.
Somehow, Autograd does not take the derivative of the numpy L1 norm.
This is likely because it is a modulus function the derivative of which
is not defined at x=0.

So, to estimate this, we have instead used the L2 norm and its derivatives.
In this example, we see whether that assumption is justified.
"""
#%%
# import numpy as np
import autograd.numpy as np
import matplotlib.pyplot as plt
from autograd import hessian


def l1_norm(params):
    """Actual L1 norm"""

    return np.linalg.norm(params, ord=1)


def l2_norm(params):
    """L2 norm"""

    return np.linalg.norm(params, ord=2)


# x = np.arange(-100, 100)
x = np.array([-0.0001, 0.15])

# Taking the Jacobians in each case:
# For L1: jac(x) = [sign(x1) sign(x2) ... sign(xn)]

# L2: 1/L2 * [x1 x2 ... xn]

# Plot the first derivative arrays
plt.figure()
plt.plot(x, np.sign(x), color="red", label=r"$\nabla_x L1 norm$")
plt.plot(x, x / l2_norm(x), color="blue", label=r"$\nabla_x L2 norm$")
plt.legend(loc="lower right")
# Taking the Hessians w.r.t parameters in each case:
# For L1 : 0_nxn

# For L2: [... d (xi/L2) / dxi dxj ...]
hessian_mat = hessian(lambda params: np.linalg.norm(params, 2))(x)

# See how different this matrix is from 0.
print("Det: " + str(np.linalg.det(hessian_mat - np.zeros((x.shape[0], x.shape[0])))))
print("Max: " + str(np.max(hessian_mat - np.zeros((x.shape[0], x.shape[0])))))

"""
Conclusions:
1. For all arrays, the first derivative of the L1 and L2 norms
    differ significantly, but the second derivatives are very close
    to each other.
2. Thus, if the objective is to compute the Hessian matrix, the
    errors with either of these approaches are exceedingly small.
"""
# %%
